import pytest

from ..smarty_streets import SmartyStreets


class MockData:
    # Mock data returned by address_lookup method
    mock_components = {"street": "1600 Penns", "city": "Cupertino", "zipcode": "95014", "state": "CA"}
    mock_metadata = {"address_data": "test_address_data"}
    mock_analysis = {"analysis_data": "test_analysis_data"}

    def __init__(self, data=None):
        if data and isinstance(data, dict):
            self.__dict__ = data
        elif data:
            self.components = data(self.mock_components)
            self.metadata = data(self.mock_metadata)
            self.analysis = data(self.mock_analysis)

    def send_lookup(self, lookup):
        assert lookup.street == "1600 Penns"
        assert lookup.state == "CA"
        lookup.result = [MockData(MockData)]

    def send(self, lookup):
        lookup.result = [MockData(self.mock_metadata)]


class TestSmartyStreets:
    @pytest.fixture
    def smarty_streets(self, monkeypatch=pytest.MonkeyPatch()):
        smarty_streets = SmartyStreets()

        # Mock the client and auto_complete_client API calls.
        monkeypatch.setattr(smarty_streets, "client", MockData())
        monkeypatch.setattr(smarty_streets, "auto_complete_client", MockData())

        return smarty_streets

    def test_address_lookup(self, smarty_streets):
        addresses = smarty_streets.address_lookup(street="1600 Penns", state="CA")

        assert addresses == [
            {
                "components": {"street": "1600 Penns", "city": "Cupertino", "zipcode": "95014", "state": "CA"},
                "metadata": {
                    "address_data": "test_address_data",
                },
                "analysis": {"analysis_data": "test_analysis_data"},
            }
        ]

    def test_auto_complete_address_lookup(self, smarty_streets):
        addresses = smarty_streets.auto_complete_address_lookup("1600 Penns")

        assert addresses == [{"address_data": "test_address_data"}]


# # TODO: def test_extract_address()



HTML_SANITIZERS = {
    "borrower": {
        "tags": {
            "h1",
            "h2",
            "h3",
            "strong",
            "em",
            "p",
            "ul",
            "ol",
            "li",
            "br",
            "sub",
            "sup",
            "hr",
            "colgroup",
            "col",
            "table",
            "div",
            "thead",
            "tr",
            "th",
            "tbody",
            "td",
        },
        "attributes": {
            "table": ("style",),
            "div": ("class",),
        },
        "empty": {"hr", "br"},
        "separate": {"p", "li", "th", "td", "tr"},
        "whitespace": {"br"},
    }
}

HTML_SANITIZER_CUSTOM_SETTINGS = {
    "tags": (
        "h1",
        "h2",
        "h3",
        "a",
        "strong",
        "em",
        "p",
        "ul",
        "ol",
        "li",
        "br",
        "hr",
        "h4",
        "table",
        "colgroup",
        "col",
        "thead",
        "tr",
        "th",
        "tbody",
        "td",
        "div",
    ),
    "attributes": {
        "a": ("href", "name", "target", "title", "id", "rel", "style"),
        "table": ("style",),
        "div": ("class",),
        "col": ("span", "style"),
    },
    "empty": {"hr", "br", "col"},
    "separate": {"p", "li", "th", "td", "tr"},
    "whitespace": {"br"},
}



import datetime
import random
import unittest

import holidays

from ..helper import (
    clean_phone_number,
    mask_phone_number,
    next_business_day,
    recursive_update,
)


class TestCleanPhoneNumber(unittest.TestCase):
    def test_clean_phone_number_us_no(self):
        assert clean_phone_number("+1(812) 4321-543") == "+1 8124321543"

    def test_clean_phone_number_ind_no(self):
        assert clean_phone_number("+91 996-788-4840") == "+91 9967884840"

    def test_clean_phone_number_no_puntuation(self):
        assert clean_phone_number("1234567890") == "1234567890"


class TestMaskPhoneNumber(unittest.TestCase):
    def test_mask_phone_number(self):
        assert mask_phone_number("1234567890") == "XXXXX67890"


class TestNextBusinessDay(unittest.TestCase):
    monday = datetime.date(2023, 3, 27)
    sunday = datetime.date(2023, 3, 26)
    thanksgiving_holiday = datetime.date(2022, 11, 24)

    def test_next_business_day_sunday(self):
        assert next_business_day(self.sunday) == self.monday

    def test_next_business_day_monday(self):
        assert next_business_day(self.monday) == self.monday

    # def test_next_business_day_holiday(self):
    #     assert next_business_day(self.thanksgiving_holiday) == datetime.date(2022, 11, 25)

    def test_next_business_day_today(self):
        date = datetime.date.today()
        us_holidays = holidays.US()
        while date.isoweekday() > 5 or date in us_holidays:
            date += datetime.timedelta(days=1)
        assert next_business_day() == date

    def test_next_business_day_holiday_random(self):
        us_holidays = holidays.US()
        rand_holiday = random.choice(list(holidays.US(years=datetime.date.today().year).keys()))
        date = rand_holiday
        while date.isoweekday() > 5 or date in us_holidays:
            date += datetime.timedelta(days=1)
        assert next_business_day(rand_holiday) == date

    def test_next_business_day_str_fail(self):
        with self.assertRaises(AttributeError):
            next_business_day("2023-03-27")


class TestRecursiveUpdate(unittest.TestCase):
    base_dict = {"name": "person", "hello": "world", "occupation": "employee"}

    def test_recursive_update_empty_dict(self):
        assert recursive_update({}, self.base_dict) == {
            "name": "person",
            "hello": "world",
            "occupation": "employee",
        }

    def test_recursive_update_contains_a_same_key_value_dict(self):
        assert recursive_update({"name": "person"}, self.base_dict) == {
            "name": "person",
            "hello": "world",
            "occupation": "employee",
        }

    def test_recursive_update_contains_a_same_key_dict(self):
        assert recursive_update({"name": "human"}, self.base_dict) == {
            "name": "person",
            "hello": "world",
            "occupation": "employee",
        }

    def test_recursive_update_contains_a_different_key_dict(self):
        assert recursive_update({"city": "new-york"}, self.base_dict) == {
            "name": "person",
            "hello": "world",
            "occupation": "employee",
            "city": "new-york",
        }

    def test_recursive_update_none_type_passed_to_be_updated(self):
        with self.assertRaises(TypeError):
            recursive_update(None, self.base_dict)

    def test_recursive_update_base_dict_none(self):
        with self.assertRaises(AttributeError):
            recursive_update({}, None)

    def test_recursive_update_both_dict_none(self):
        with self.assertRaises(AttributeError):
            recursive_update(None, None)



import logging
from uuid import uuid4

import sentry_sdk
from django.conf import settings
from django.core.exceptions import ValidationError as CoreValidationError
from django.db import IntegrityError
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy as _
from django_error_reporting.utils import add_event_tag
from rest_framework import status
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.fields import get_error_detail
from rest_framework.response import Response
from rest_framework.views import exception_handler

LOGGER = logging.getLogger("root")


def fmt_error_msg(
    err_code="F1111", usr_msg=None, tech_msg=None, instance_id=None, exc_obj=None, display_uniqueid=False
):
    uniqueid = str(uuid4())  # this will be used to filter in Sentry
    instance_id = str(instance_id) if instance_id is not None else ""

    if usr_msg is None:
        usr_msg = _("An unspecified error has occurred.  No further details at this time.")
        display_uniqueid = True
    elif not isinstance(usr_msg, str) and not isinstance(usr_msg, Promise):
        usr_msg = _("An invalid error message format was received.  No further details at this time.")
        display_uniqueid = True

    msg_to_display = f"{err_code} - {usr_msg}"
    if display_uniqueid is True:
        msg_to_display = f"{err_code} - {usr_msg}  -  ID: {uniqueid}"

    if tech_msg is None:
        if isinstance(usr_msg, str):
            tech_msg = usr_msg
        else:
            tech_msg = "No specific technical details are available."

    if tech_msg != "" and isinstance(tech_msg, str):
        sentry_msg = f"{tech_msg} - Instance: {instance_id}" if instance_id != "" else tech_msg

    else:
        sentry_msg = "tech_msg is either missing or a non-string object.  No further details at this time."

    add_event_tag("err_code", err_code)
    add_event_tag("uniqueId", uniqueid)

    sentry_sdk.capture_message(sentry_msg)

    if exc_obj is not None:
        sentry_sdk.capture_exception(exc_obj)

    return msg_to_display


def flatten_drf_validations(errors, path=None):
    if path:
        path = path.replace("_", " ").title()

    errors_list = []
    if isinstance(errors, str):
        return [f"{path} - {errors}" if path else errors]

    if isinstance(errors, list):
        for error in errors:
            errors_list = errors_list + flatten_drf_validations(error, path=path)
        return errors_list

    if isinstance(errors, dict):
        for key, val in errors.items():
            new_path = f"{path} - {key}" if path else key
            errors_list = errors_list + flatten_drf_validations(val, path=new_path)
        return errors_list


class CustomError(APIException):
    status_code = 400
    default_detail = "An error has occurred. Please validate your submission values."

    def __init__(self, *args, **kwargs):
        super().__init__(*args)

        if "status_code" in kwargs:
            self.status_code = kwargs["status_code"]


class Http404Exception(APIException):
    status_code = 404
    default_detail = "Object does not exist."


class PermissionDeniedError(APIException):
    status_code = 403
    default_detail = "You do not have permission to perform this action"


def APIExceptionHandler(exc, context):
    # handle django core ValidationErrors
    if isinstance(exc, CoreValidationError):
        if hasattr(exc, "message_dict"):
            detail = exc.message_dict
        elif hasattr(exc, "message"):
            detail = exc.message
        elif hasattr(exc, "messages"):
            detail = exc.messages
        else:
            detail = get_error_detail(exc)

        exc = ValidationError(detail=detail)

    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)
    if response and hasattr(response, "data") and isinstance(response.data, dict):
        response.data = flatten_drf_validations(response.data)

    try:
        uid = context["request"].session["trace_id"]
    except Exception:
        uid = uuid4()

    # if there is an IntegrityError and the error response hasn't already been generated
    if isinstance(exc, IntegrityError) and not response:
        response = Response({"message": str(exc)}, status=status.HTTP_400_BAD_REQUEST)
    # If its an unhandled exception (ie 500) show a proper json response with a trace id
    if response is None and settings.DEBUG is False:
        response = Response(
            {"trace_id": str(uid), "error": True, "content": "Something went wrong"},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )

    try:
        sentry_sdk.capture_exception(exc)
    except Exception as excp:
        LOGGER.exception(f"Error: [{excp}]")

    return response



import traceback
import uuid
import warnings
from collections import OrderedDict

import semantic_version
import structlog
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import EmptyResultSet, ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import ManyToOneRel, OneToOneRel, QuerySet
from django.db.models.fields import (
    BigIntegerField,
    DecimalField,
    FloatField,
    IntegerField,
    PositiveSmallIntegerField,
)
from django.db.models.fields.json import JSONField
from django.db.models.fields.related import ForeignKey
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from model_utils.managers import QueryManager
from rest_framework import serializers
from rest_framework.fields import SkipField
from rest_framework.relations import PKOnlyObject
from semantic_version.django_fields import VersionField

from los.utils.sentry import capture_message

# from los.auditlog.models import LogEntry
from .config.exceptions import fmt_error_msg

STRUCTLOGGER = structlog.get_logger(__name__)


class InvalidOperation(Exception):
    pass


class SlightlySaferSoftDeletableQuerySetMixin:
    """
    QuerySet for SoftDeletableModel. Instead of removing instance sets
    its ``is_removed`` field to True.
    """

    def delete(self):
        """
        Soft delete objects from queryset (set their ``is_removed``
        field to True)
        """
        # TODO: uncomment this statement when ready to deprecate
        # raise InvalidOperation("Calling delete here is not allowed! Use `bulk_delete` on the Model")

        warnings.warn(
            "Calling delete is going to be deprecated in future! Use `bulk_delete` on the Model class",
            DeprecationWarning,
        )

        return self.unsafe_delete()

    def unsafe_delete(self):
        """
        Soft delete objects from queryset (set their ``is_removed``
        field to True)
        """
        tb = str(traceback.extract_stack())
        try:
            qry = str(self.query)
        except EmptyResultSet:
            qry = "***** EMPTY QUERYSET *****"

        from los.requests.tasks import create_bulk_delete_audit_event
        from los.utils.celery import queue_task

        content_type = ContentType.objects.get_for_model(self.model)
        content_type_id = content_type.id
        object_ids = list(map(str, self.values_list("id", flat=True)))

        if not object_ids:
            return 0

        STRUCTLOGGER.info(
            "Queueing the celery task for creating bulk delete audit events",
            object_ids=object_ids,
            content_type_id=content_type_id,
        )

        queue_task(
            create_bulk_delete_audit_event,
            object_ids=object_ids,
            content_type_id=content_type_id,
            queue="audit_log_queue",
            countdown=3,
            traceback_stack=tb,
            str_queryset=qry,
        )

        # still doing this operation here because many of our APIs do a bulk delete followed by a DB call to do other updates
        fields_to_update = {"is_removed": True}

        if hasattr(content_type, "modified"):
            fields_to_update["modified"] = timezone.now()

        result = self.update(**fields_to_update)

        # TODO: revisit this in future and have some stream setup rather than sentry exceptions
        capture_message(
            "Processed a bulk soft delete operation on Model: [{}]. First object ID: [{}]".format(
                content_type.app_labeled_name, object_ids[0]
            )
        )

        return result


class SlightlySaferSoftDeletableQuerySet(SlightlySaferSoftDeletableQuerySetMixin, QuerySet):
    pass


class SlightlySaferSoftDeletableManagerMixin:
    """
    Manager that limits the queryset by default to show only not removed
    instances of model.
    """

    _queryset_class = SlightlySaferSoftDeletableQuerySet

    def __init__(self, *args, _emit_deprecation_warnings=False, **kwargs):
        self.emit_deprecation_warnings = _emit_deprecation_warnings
        super().__init__(*args, **kwargs)

    def get_queryset(self):
        """
        Return queryset limited to not removed entries.
        """

        if self.emit_deprecation_warnings:
            warning_message = (
                "{0}.objects model manager will include soft-deleted objects in an "
                "upcoming release; please use {0}.available_objects to continue "
                "excluding soft-deleted objects. See "
                "https://django-model-utils.readthedocs.io/en/stable/models.html"
                "#softdeletablemodel for more information."
            ).format(self.model.__class__.__name__)
            warnings.warn(warning_message, DeprecationWarning)

        kwargs = {"model": self.model, "using": self._db}
        if hasattr(self, "_hints"):
            kwargs["hints"] = self._hints

        return self._queryset_class(**kwargs).filter(is_removed=False)


class SlightlySaferSoftDeletableManager(SlightlySaferSoftDeletableManagerMixin, models.Manager):
    pass


class SoftDeleteMixin(models.Model):
    is_removed = models.BooleanField(default=False, editable=False)

    class Meta:
        abstract = True

    objects = SlightlySaferSoftDeletableManager()
    deleted = QueryManager(is_removed=True)
    all_objects = models.Manager()

    def save(self, *args, **kwargs):
        if (self._state.adding or self.pk is None) and hasattr(self, "created"):
            # reset the created date if it's already set for a new object being added
            self.created = timezone.now()

        super().save(*args, **kwargs)

    def delete(self, using=None, soft=True, *args, **kwargs):
        """
        Soft delete object (set its ``is_removed`` field to True).
        Actually delete object if setting ``soft`` to False.
        """
        if soft:
            self.is_removed = True
            self._force_save = True  # pass _force_save in case the model supports it
            self.save(using=using)
        else:
            return super().delete(using=using, *args, **kwargs)

    @classmethod
    def bulk_delete(cls, **kwargs):
        """Slightly safer bulk soft delete.  This method attempts to ensure that no argument passed to it is `None`.
        Instead, if you really intend to filter by `None`, you should use `field__isnull=True`.

        We'll probably want to do this for other methods like `filter` and `exclude` as well, because we have run into
        scenarios where filtering on the value of a nullable FK from another model has caused some bad results too

        will also need a way to exclude fields

        kwargs only allowed for now...no Q objects, but we should be able to call Q.deconstruct() and loop through
        any child Q objects to determine if we received a `field = None` query
        """
        if not kwargs:
            raise InvalidOperation("Filters cannot be blank")

        for key, value in kwargs.items():
            if value is None:
                raise InvalidOperation("Cannot filter by None.  Must use `__isnull=True`")

        STRUCTLOGGER.info("Executing a bulk soft delete operation", klass=cls, arguments=kwargs)
        cls.objects.filter(**kwargs).unsafe_delete()  # nosemgrep


class VersionMixin(models.Model):
    version = VersionField(editable=False, default="1.0.0")
    version_major = models.IntegerField(editable=False, null=True)
    version_minor = models.IntegerField(editable=False, null=True)
    version_patch = models.IntegerField(editable=False, null=True)

    class Meta:
        abstract = True

    # current_version = QueryManager(is_removed=False)
    # deleted = QueryManager(is_removed=True)
    # all_objects = models.Manager()

    def save(self, *args, **kwargs):
        if isinstance(self.version, str):
            self.version = semantic_version.Version(self.version)
        self._sync_version_parts()
        return super().save(*args, **kwargs)

    def _sync_version_parts(self):
        if self.version:
            self.version_major = self.version.major
            self.version_minor = self.version.minor
            self.version_patch = self.version.patch


class JsonSerializeMixin:
    def model_to_dict(self, return_fk_as_id=False):
        serialized_model = {}
        for field in self._meta.fields:
            value = getattr(self, field.name)
            if (
                value is False
                or value is True
                or value is None
                or type(field) in [BigIntegerField, FloatField, IntegerField, PositiveSmallIntegerField]
            ):
                serialized_model[field.name] = value
            elif type(field) == DecimalField:
                serialized_model[field.name] = float(value) if value else None
            elif type(field) == ForeignKey and return_fk_as_id:
                if isinstance(value.pk, uuid.UUID):
                    serialized_model[f"{field.name}_id"] = str(value.pk)
                else:
                    serialized_model[f"{field.name}_id"] = value.pk
            elif isinstance(field, JSONField):
                serialized_model[field.name] = value
            else:
                serialized_model[field.name] = str(value)

        for relation in self._meta.related_objects:
            if hasattr(relation.related_model, "model_to_dict"):
                serialized_model[relation.name] = []
                if hasattr(self, relation.name):
                    if type(relation) == OneToOneRel:
                        serialized_model[relation.name] = getattr(self, relation.name).model_to_dict()
                    # elif type(getattr(self, relation.name)) == ManyToOneRel:
                    elif type(relation) == ManyToOneRel:
                        for obj in getattr(self, relation.name).all():
                            serialized_model[relation.name].append(obj.model_to_dict())

        return serialized_model


# TODO: THIS USES THE LOS AUDIT LOG MODELS, SO LOOK AT HOW COULD WE IMPLEMENT THIS.

# class LoggingMixin:
#     def save_and_log(self, request=None, action=LogEntry.Action.UPDATE, log_as_task=False, extras=None, **kwargs):
#         if extras is None:
#             extras = {}
#         changes = {}
#         for key, value in kwargs.items():
#             if str(getattr(self, key)) != str(value):
#                 changes[key] = [str(getattr(self, key)), str(value)]
#                 setattr(self, key, value)

#         if request:
#             actor = request.user
#             ip = get_client_ip(request)
#         else:
#             actor = None
#             ip = None

#         if extras:
#             changes.update(extras)

#         self.save()
#         if log_as_task:
#             LogEntry.objects.log_create_as_task(
#                 instance=self,
#                 action=action,
#                 actor=actor,
#                 changes=changes,
#                 remote_addr=ip,
#             )
#         else:
#             LogEntry.objects.log_create(
#                 instance=self,
#                 action=action,
#                 actor=actor,
#                 changes=changes,
#                 remote_addr=ip,
#             )


class MutuallyExclusiveSerializerMixin(serializers.Serializer):
    """The Highlander of serializer validators"""

    MUTUALLY_EXCLUSIVE_FIELDS = []
    REQUIRED = "required"  # Required means one or the other of the fields is required
    OPTIONAL = "optional"  # Optional means the field is optional, but if provided, only one can be set
    EXCLUSIVITY_TYPE = REQUIRED

    def validate(self, attrs, *args, **kwargs):
        if self.MUTUALLY_EXCLUSIVE_FIELDS:
            populated = 0
            for field in self.MUTUALLY_EXCLUSIVE_FIELDS:
                if attrs.get(field, None):
                    populated += 1

            if self.EXCLUSIVITY_TYPE == self.REQUIRED:
                if populated != 1:
                    err_code = "E1092"
                    usr_msg = _("Exactly one of the following fields must be set: {}").format(
                        self.MUTUALLY_EXCLUSIVE_FIELDS
                    )
                    tech_msg = f"Exactly one of the following fields must be set: {self.MUTUALLY_EXCLUSIVE_FIELDS}"
                    error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
                    raise ValidationError(error_msg)
            elif self.EXCLUSIVITY_TYPE == self.OPTIONAL:
                if populated > 1:
                    err_code = "E1093"
                    usr_msg = _("Only one of the following fields can be set: {}").format(
                        self.MUTUALLY_EXCLUSIVE_FIELDS
                    )
                    tech_msg = f"Only one of the following fields can be set: {self.MUTUALLY_EXCLUSIVE_FIELDS}"
                    error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
                    raise ValidationError(error_msg)

        return super().validate(attrs, *args, **kwargs)


class DataMaskingSerializerMixin(serializers.Serializer):
    """
    Mask PII or other sensitive information
        - mask_field - Hook to write custom masking function if needed
        - unmask_fields - Hook to write custom permissions if needed. DEFAULT: Mask sensitive fields for loans in non-editable status.
          NOTE: Request context must be passed to serializer to check for user perms
    """

    class Meta:
        SENSITIVE_FIELDS = []

    def __init__(self, instance=None, *args, **kwargs):
        super().__init__(instance, *args, **kwargs)
        self.UNMASK_FIELDS = self.unmask_fields(instance, *args, **kwargs)

    def unmask_fields(self, instance, *args, **kwargs):
        """
        Hook to add custom logic for any scenarios where the fields need to be unmasked.
        NOTE: Request context must be passed to serializer to check for user perms
        """
        return bool(kwargs.get("context", {}).get("unmask_fields", False))

    def mask_field(self, value, *args, **kwargs):
        """Hook to write custom masking function if needed"""
        return self.dynamic_mask_account_number(value)

    @staticmethod
    def dynamic_mask_account_number_last_four(account_number):
        if not account_number:
            return account_number
        account_number = str(account_number).strip()
        return account_number[:-4] + "XXXX"

    @staticmethod
    def dynamic_mask_account_number(account_number):
        if not account_number:
            return account_number
        account_number = str(account_number)
        length = len(account_number)
        masked = ""
        if length > 4:
            last_displayed_digits = 4
        else:
            last_displayed_digits = 1

        last_digits = account_number[-last_displayed_digits:]
        for i in range(length - last_displayed_digits):
            if account_number[i].isalnum():
                masked += "X"
            else:
                masked += account_number[i]  # keep delimiter(s)
        return masked + last_digits

    @staticmethod
    def mask_account_number(account_number):
        if account_number is None:
            return account_number
        account_number = str(account_number)
        last_four = account_number[-4:]
        masked = ""
        for i in range(len(account_number) - 4):
            if account_number[i].isalnum():
                masked += "X"
            else:
                masked += account_number[i]  # keep delimiter(s)
        return masked + last_four

    def to_representation(self, instance):
        ret = OrderedDict()
        fields = self._readable_fields

        for field in fields:
            try:
                attribute = field.get_attribute(instance)
            except SkipField:
                continue

            check_for_none = attribute.pk if isinstance(attribute, PKOnlyObject) else attribute
            if check_for_none is None:
                ret[field.field_name] = None
            else:
                value = field.to_representation(attribute)
                if self.UNMASK_FIELDS is False and field.field_name in self.Meta.SENSITIVE_FIELDS:
                    value = self.mask_field(value, field_name=field.field_name)
                ret[field.field_name] = value
        return ret


class DontUpdateModelSerializerMixin:
    @staticmethod
    def needs_saving(data_dict, instance_dict):
        for k, v in data_dict.items():
            if k in instance_dict and instance_dict[k] != v:
                return True
        return False

    def dont_update(self, instance, validated_data):
        # same as update() except no instance.save()
        # serializers.raise_errors_on_nested_writes('update', self, validated_data)
        # info = model_meta.get_field_info(instance)

        # m2m_fields = []
        # for attr, value in validated_data.items():
        #     if attr in info.relations and info.relations[attr].to_many:
        #         m2m_fields.append((attr, value))
        #     else:
        #         setattr(instance, attr, value)

        # for attr, value in m2m_fields:
        #     field = getattr(instance, attr)
        #     field.set(value)

        return instance


class CopyMixin(models.Model):
    copied_from = models.ForeignKey("self", on_delete=models.PROTECT, null=True, blank=True, related_name="copies")
    copied_data = models.JSONField(
        "JSON of values copied from Original", blank=True, null=True, encoder=DjangoJSONEncoder
    )

    class Meta:
        abstract = True


class ProtectedFieldsModelMixin:
    @cached_property
    def old_version(self):
        return self.__class__.objects.filter(pk=self.pk).first()

    def is_protected_fields_modified(self, protected_fields, interface, return_modified_fields: bool = False):
        if not interface:
            if return_modified_fields:
                return True, []

            return True

        return interface.parameters_changed(self.get_protected_fields_values(protected_fields), return_modified_fields)

    def get_protected_fields_values(self, protected_fields):
        return {field: getattr(self, field) for field in protected_fields}

    def validate_interfaces(self, *args, **kwargs):
        raise NotImplementedError


# class ReadReplicaMixin:
#     """
#     ModelViewSet mixin to force list/retrieve methods to the replica db
#     """

#     def use_read_replica(self, request, *args, **kwargs) -> bool:
#         """
#         Determines if queries in the scope of the given request
#         should use the read replica.
#         Args:
#             request (Request): The HTTP request
#         Returns:
#             bool: Whether this request should use the read replica.
#         """
#         return request.method.lower() == "get"

#     def dispatch(self, request, *args, **kwargs):
#         """Builtin dispatch method for a DRF ModelViewSet.
#         This is where we are overriding all GET methods to use the read replica.

#         Args:
#             request (HttpRequest): request object

#         Returns:
#             response (HttpResponse, JsonResponse): Response object
#         """
#         new_request = self.initialize_request(request, *args, **kwargs)
#         if self.use_read_replica(new_request, *args, **kwargs):
#             with use_replica_db:
#                 return super().dispatch(request, *args, **kwargs)
#         return super().dispatch(request, *args, **kwargs)



def import_dotted(path: str):
    path_components = path.split(".")
    obj = path_components[-1]
    fromlist = [obj]
    mod = __import__(".".join(path_components[0:-1]), fromlist=fromlist)
    return getattr(mod, obj)



import json
import unittest

import semantic_version

from ..encoders import VersionCompatibleEncoder


class TestVersionCompatibleEncoder(unittest.TestCase):
    def test_version_compatible_encoder_pass(self):
        data = {"version": semantic_version.Version("1.2.3")}
        encoded_data = json.dumps(data, cls=VersionCompatibleEncoder)
        decoded_data = json.loads(encoded_data)
        self.assertEqual(decoded_data["version"], "1.2.3")

    def test_version_compatible_encoder_fail(self):
        data = {"version": semantic_version.Version("1.2.3")}
        encoded_data = json.dumps(data, cls=VersionCompatibleEncoder)
        decoded_data = json.loads(encoded_data)
        self.assertNotEqual(decoded_data["version"], semantic_version.Version("1.2.3"))

    def test_version_compatible_encoder_type_str_pass(self):
        data = {"version": semantic_version.Version("1.2.3")}
        encoded_data = json.dumps(data, cls=VersionCompatibleEncoder)
        decoded_data = json.loads(encoded_data)
        self.assertIsInstance(decoded_data["version"], str)

    def test_version_compatible_encoder_version_instance_fail(self):
        data = {"version": semantic_version.Version("1.2.3")}
        encoded_data = json.dumps(data, cls=VersionCompatibleEncoder)
        decoded_data = json.loads(encoded_data)
        self.assertNotIsInstance(decoded_data["version"], semantic_version.Version)



from ..read_replica import (
    pin_this_replica_thread,
    pin_this_thread,
    this_replica_thread_is_pinned,
    this_thread_is_pinned,
    unpin_this_replica_thread,
    unpin_this_thread,
    use_default_db,
    use_replica_db,
)


class TestReadReplica:
    def test_use_default_db(self):
        @use_default_db
        def test_func():
            assert this_thread_is_pinned() is True

        test_func()
        assert this_thread_is_pinned() is False

    def test_use_replica_db(self):
        @use_replica_db
        def test_func():
            assert this_replica_thread_is_pinned() is True

        test_func()
        assert this_replica_thread_is_pinned() is False

    def test_pin_this_thread(self):
        pin_this_thread()
        assert this_thread_is_pinned() is True

    def test_unpin_this_thread(self):
        pin_this_thread()
        assert this_thread_is_pinned() is True
        unpin_this_thread()
        assert this_thread_is_pinned() is False

    def test_pin_this_replica_thread(self):
        pin_this_replica_thread()
        assert this_replica_thread_is_pinned() is True

    def test_unpin_this_replica_thread(self):
        pin_this_replica_thread()
        assert this_replica_thread_is_pinned() is True
        unpin_this_replica_thread()
        assert this_replica_thread_is_pinned() is False



import phonenumbers
from phonenumber_field.phonenumber import PhoneNumber
from phonenumber_field.serializerfields import PhoneNumberField as BasePhoneNumberField
from rest_framework import serializers


class DynamicFieldsModelSerializer(serializers.ModelSerializer):
    """
    A ModelSerializer that takes an additional `fields` and/or `exclude` argument that
    control which fields should be displayed (or not). Also supports double underscore
    notation for nested serializers.

    Excludes take precedence if a field is present in both `fields` and `exclude`

    Example Usage:

    serializer = DynamicFieldsModelSerializer(
        fields=['id', 'user', 'name', 'updated'],
        exclude=['user__address']
    )

    This would only include `id`, `user`, `name`, `updated`, and `user.address` would be removed
    """

    def __init__(self, *args, **kwargs):
        def parse_nested_fields(fields, exclude=False):
            field_object = {"fields": []}
            for f in fields:
                obj = field_object
                nested_fields = f.split("__")
                for v in nested_fields:
                    if v not in obj["fields"]:
                        obj["fields"].append(v)
                    if nested_fields.index(v) < len(nested_fields) - 1:
                        obj[v] = obj.get(v, {"fields": []})
                        obj = obj[v]
            return field_object

        def select_nested_fields(serializer, fields, exclude=False):
            for k in fields:
                if str(k).lower() == "fields":
                    if exclude:
                        fields_to_exclude(serializer, fields[k], fields.keys())
                    else:
                        fields_to_include(serializer, fields[k])
                else:
                    select_nested_fields(serializer.fields[k], fields[k], exclude)

        def fields_to_include(serializer, fields):
            # Drop any fields that are not specified in the `fields` argument.
            allowed = set(fields)
            existing = set(serializer.fields.keys())
            for field_name in existing - allowed:
                serializer.fields.pop(field_name, None)

        def fields_to_exclude(serializer, fields, ignore):
            # Drop any fields that ARE specified in the `fields` argument.
            for field_name in fields:
                if field_name not in ignore:
                    serializer.fields.pop(field_name, None)

        # Don't pass the 'fields' arg up to the superclass
        fields = kwargs.pop("fields", None)
        exclude = kwargs.pop("exclude", None)

        # Instantiate the superclass normally
        super().__init__(*args, **kwargs)

        if fields is not None:
            fields = parse_nested_fields(fields)
            select_nested_fields(self, fields)

        if exclude is not None:
            exclude = parse_nested_fields(exclude, exclude=True)
            select_nested_fields(self, exclude, exclude=True)


class PhoneNumberField(BasePhoneNumberField):
    def __init__(self, format=phonenumbers.PhoneNumberFormat.E164, **kwargs):
        self.format = format
        super().__init__(**kwargs)

    def to_representation(self, value: PhoneNumber):
        if not value:
            return None

        return value.format_as(self.format)



from drf_spectacular.openapi import AutoSchema


class LosApiSchema(AutoSchema):
    def get_tags(self):
        from config import api_router, api_router_public

        _tags = ["api"]
        for prefix, viewset, basename in api_router.router.registry:
            if viewset.__name__ == self.view.__class__.__name__:
                _tags.append("tenant")
                break
        for prefix, viewset, basename in api_router_public.router.registry:
            if viewset.__name__ == self.view.__class__.__name__:
                _tags.append("public")
                break

        return _tags



import collections.abc
import datetime

import holidays


def clean_phone_number(phone_number):
    """
    Converts phone number from the format +1(812) 4321-543 into the format +1 8124321543.

    Args:
        phone_number (str): A phone number in the format +1(812) 4321-543.

    Returns:
        str: The phone number in the format +1 8124321543.
    """
    for r in (("(", " "), (") ", ""), ("-", "")):
        phone_number = phone_number.replace(*r)
    return phone_number


def mask_phone_number(phone_number):
    """
    Masks a phone number by replacing the first five digits with 'X'.

    Args:
        phone_number (str): The phone number to be masked.

    Returns:
        str: The masked phone number.
    """
    if phone_number is None:
        return phone_number
    phone_number = str(phone_number)
    length = len(phone_number)
    masked = ""

    last_digits = phone_number[5:]
    for i in range(length - 5):
        if phone_number[i].isalnum():
            masked += "X"
        else:
            masked += phone_number[i]  # keep delimiter(s)
    return masked + last_digits


def recursive_update(d, u):
    """
    Recursively updates a dictionary with another dictionary.

    Args:
        d (dict): The dictionary to be updated.
        u (dict): The dictionary to update with.

    Returns:
        dict: The updated dictionary.
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def next_business_day(date=None):
    """
    Returns the next business day from the given date.

    Args:
        date (datetime.date, optional): The date to start from. Defaults to today.

    Returns:
        datetime.date: The next business day.
    """
    if date is None:
        date = datetime.date.today()

    us_holidays = holidays.US()
    while date.isoweekday() > 5 or date in us_holidays:
        date += datetime.timedelta(days=1)
    return date


def tz_aware(dt):
    """
    Checks if a datetime object is timezone aware.

    Args:
        dt (datetime.datetime): The datetime object to check.

    Returns:
        bool: True if the datetime object is timezone aware, False otherwise.
    """
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None



import json
import re
from io import BytesIO

from html_sanitizer import Sanitizer
from html_sanitizer.django import get_sanitizer
from openpyxl import load_workbook

from .sanitizers import HTML_SANITIZER_CUSTOM_SETTINGS


def xlsx_to_json(xlsx_file, group_by_key=None):
    wb = load_workbook(BytesIO(xlsx_file.read()))
    # Iterate through each row in worksheet and fetch values into dict
    sheet_obj = wb.active

    # Cell objects also have a row, column,
    # and coordinate attributes that provide
    # location information for the cell.

    # Note: The first row or
    # column integer is 1, not 0.
    # cell_obj = sheet_obj.cell(row=1, column=1)

    group_key_idx = None
    HEADERS = []
    for col_idx in range(1, sheet_obj.max_column + 1):
        data = sheet_obj.cell(row=1, column=col_idx).value
        HEADERS.append(data)
        if group_by_key == data:
            group_key_idx = col_idx

    if group_key_idx:
        data_list = {}
    else:
        data_list = []

    for row_idx in range(2, sheet_obj.max_row + 1):
        row_obj = {}
        key = sheet_obj.cell(row=row_idx, column=group_key_idx).value if group_key_idx else None

        for header_idx, col_idx in enumerate(range(1, sheet_obj.max_column + 1)):
            if col_idx != group_key_idx:
                row_obj[HEADERS[header_idx]] = sheet_obj.cell(row=row_idx, column=col_idx).value

        if group_key_idx:
            if key not in data_list:
                data_list[key] = []
            data_list[key].append(row_obj)
        else:
            data_list.append(row_obj)

    # Serialize the list of dicts to JSON
    return json.dumps(data_list)


# TODO: Need the path as args for this so that it can be reused accross projects
# def html_to_pdf(html, stylesheets=None):
#     bio = BytesIO()
#     ROOT_DIR = Path(__file__).resolve(strict=True).parent.parent.parent
#     APPS_DIR = str(ROOT_DIR / "los/templates")
#     if stylesheets is None:
#         stylesheets = [CSS(APPS_DIR + "/css/los_app.css")]
#     else:
#         stylesheets = [CSS(APPS_DIR + f"/css/{stylesheets}")]

#     kwargs = {"string": html}
#     # if stylesheets:
#     #     kwargs["stylesheets"] = stylesheets

#     HTML(**kwargs).write_pdf(bio, stylesheets=stylesheets)
#     bio.seek(0)
#     return bio


# NOTE: imported a library inside a function, why so?
def convert_numbers_to_words(number):
    import inflect

    p = inflect.engine()

    number_words = p.number_to_words(number)
    return number_words


def parse_json_string(value):
    try:
        value = json.loads(value)
    except json.decoder.JSONDecodeError:
        value
    return value if value != "-" else None


def snake_case(s: str) -> str:
    """Utility function to convert strings in CamelCase to snake_case

    Args:
        s (str): input string

    Returns:
        str: snake_case string
    """
    s = re.sub("[^0-9a-zA-Z]+", "", s)
    return "_".join(re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", s.replace("-", " "))).split()).lower()


def phone_format(number: str) -> str:
    """Format phone number with dashes

    Args:
        number (str): unformatted phone number

    Returns:
        str: formatted phone number ie 555-555-5555
    """
    number = re.sub("[^0-9]", "", number)
    number = number[-10:]
    number = f"{number[:3]}-{number[3:6]}-{number[6:]}"
    return number


def clean_html(value: str, sanitizer="default") -> str:
    """Helper function to sanitize HTML to prevent cross-site scripting vulnerabilities.  If no `sanitizer`
    is specified, we'll default to the "default" sanitizer, and pass in our own custom settings from
    `settings.HTML_SANITIZER_CUSTOM_SETTINGS`

    Args:
        value (str): input html string
        sanitizer (str, optional): Sanitizer name to pass to `get_sanitizer` function. Defaults to "default"

    Returns:
        str: cleaned html string
    """
    if not isinstance(value, str):
        return value
    if sanitizer == "default":
        # use the default serializer with custom settings
        sanitize_obj = Sanitizer(HTML_SANITIZER_CUSTOM_SETTINGS)
    else:
        # use the specified serializer
        sanitize_obj = get_sanitizer(name=sanitizer)

    return sanitize_obj.sanitize(value)



from urllib.parse import urlparse

from django import template
from django.urls import reverse

register = template.Library()


@register.simple_tag(takes_context=True)
def tenant_url(context, view_name, *args, tenant=None, **kwargs):
    """
    Custom template tag to allow for redirects between schemas (ie; for the globalsuperuser dashboard). Django's native urlresolvers strip out subdomains.
    Can be customized for other use cases in the future
    """
    request = context["request"]
    absolute_uri = request.build_absolute_uri(reverse(view_name, args=args, kwargs=kwargs))
    if tenant:
        current_tenant = request.tenant if hasattr(request, "tenant") else None
        parsed_uri = urlparse(absolute_uri)
        constructed_netloc = (
            parsed_uri.netloc.replace(current_tenant, tenant)
            if current_tenant and current_tenant in parsed_uri.netloc
            else f"{tenant}.{parsed_uri.netloc}"
        )
        absolute_uri = f"{parsed_uri.scheme}://{constructed_netloc}{parsed_uri.path}"

    return absolute_uri



import phonenumbers
from django import template
from django.conf import settings

register = template.Library()


@register.simple_tag()
def formatted_phone_number(phone_number, format=None, region=None):
    """Template tag to format a phone number in various formats.

    Args:
        phone_number (str | PhoneNumber): The phone number to format.
        format (str, optional): The target phone number format, one of "E164", "INTERNATIONAL", "NATIONAL", "RFC3966".
            Defaults to None, which will use the application default.
        region (str, optional): The region of the source phone number, if not in E164 format.
            Defaults to None, which will use the application default.

    Returns:
        str: The phone number in the target format.
    """
    if not phone_number:
        return None

    if not format:
        format = getattr(settings, "PHONENUMBER_DEFAULT_FORMAT", None)
    else:
        format = getattr(phonenumbers.PhoneNumberFormat, format, None)

    if not format:
        return phone_number

    if not region:
        region = getattr(settings, "PHONENUMBER_DEFAULT_REGION", None)

    if not isinstance(phone_number, phonenumbers.PhoneNumber):
        phone_number = str(phone_number)
        phone_number = phonenumbers.parse(number=phone_number, region=region)

    return phonenumbers.format_number(phone_number, format)



from whitenoise.storage import CompressedManifestStaticFilesStorage


class OriginalWebpackOutput(CompressedManifestStaticFilesStorage):
    """
    Customized whitenoise storage system that ignores hashing of webpack output.

    To integrate add the following to your configuration:
    STATICFILES_STORAGE = "los.utils.static_webpack.OriginalWebpackOutput"
    """

    def hashed_name(self, name, content=None, filename=None):
        """
        Do not hash webpack output.
        """
        if name.startswith("final/"):
            if self._new_files is not None:
                self._new_files.add(self.clean_name(name))
            return name
        else:
            return super().hashed_name(name, content, filename)



from django import template
from django.conf import settings
from django.db import connection
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def SENTRY_SCRIPTS_INCLUDE():
    html = ""
    if (
        hasattr(settings, "SENTRY_DSN")
        and settings.SENTRY_DSN
        and "https://" in settings.SENTRY_DSN
        and "@" in settings.SENTRY_DSN
    ):
        js_dsn = f"https://js.sentry-cdn.com/{settings.SENTRY_DSN.replace('https://','').split('@')[0]}.min.js"
        html = f'<script src="{js_dsn}" crossorigin="anonymous"></script>'
    return mark_safe(html)  # nosemgrep


@register.simple_tag
def SENTRY_TAGS_INCLUDE(user=None):
    html = ""

    if (
        hasattr(settings, "SENTRY_DSN")
        and settings.SENTRY_DSN
        and "https://" in settings.SENTRY_DSN
        and "@" in settings.SENTRY_DSN
    ):
        release = settings.CODEBUILD_BUILD_NUMBER or "Local-Build"
        base_url = settings.BASE_URL
        commit_id = settings.CODEBUILD_RESOLVED_SOURCE_VERSION or "Local-Build"
        branch_name = settings.CODEBUILD_WEBHOOK_HEAD_REF or "Local-Branch"
        schema = connection.schema_name
        email = user.email if user and user.is_authenticated else ""
        user_id = user.id if user else ""
        username = user.username if user and user.is_authenticated else ""
        schema_user_id = f"{schema}:{user_id}" if schema and user_id else ""
        schema_user_name = f"{schema}:{username}" if schema and username else ""

        html = f"""
    <script>
    Sentry.onLoad(function() {{
        Sentry.init({{
        release: '{release}',
        environment: '{base_url}'
        }});
        Sentry.configureScope(scope => {{
        scope.setTag( 'build_number', '{release}' );
        scope.setTag( 'build_commit', '{commit_id}' );
        scope.setTag( 'build_branch', '{branch_name}' );
        scope.setTag( 'schema', '{schema}' );
        scope.setTag( 'email', '{email}' );
        scope.setTag( 'schema_user_id', '{schema_user_id}' );
        scope.setTag( 'schema_user_name', '{schema_user_name}' );
        }});
    }});
    </script>    """

    return mark_safe(html.replace("\n", " "))  # nosemgrep



import json

from django.utils.translation import gettext_lazy as _

from ..config.exceptions import fmt_error_msg


class ErrorCode:
    __slots__ = ("data", "error_codes_path")

    def __init__(self, error_codes_file_path):
        self.error_codes_path = error_codes_file_path

        with open(self.error_codes_path) as file:
            self.data = json.loads(file.read())

    def error_message(self, error_code, data=None, exception_object=None, display_uniqueid=False, instance_id=None):
        user_message = self.data[error_code]["usr_msg"]
        tech_message = self.data[error_code]["tech_msg"]

        if data:
            user_message = user_message.format(**data)
            tech_message = tech_message.format(**data)

        translated_user_message = _(user_message)

        return fmt_error_msg(
            err_code=error_code,
            usr_msg=translated_user_message,
            tech_msg=tech_message,
            exc_obj=exception_object,
            display_uniqueid=display_uniqueid,
            instance_id=instance_id,
        )



from django.contrib.auth import get_user_model
from factory import Faker, LazyAttribute, fuzzy, post_generation
from factory.django import DjangoModelFactory


class UserFactory(DjangoModelFactory):
    username = LazyAttribute(
        lambda obj: "{}_{}".format(obj.name.lower().replace(" ", ""), fuzzy.FuzzyInteger(0, 99999).fuzz())
    )
    email = LazyAttribute(lambda obj: "%s@gmail.com" % (obj.username))
    name = Faker("name")

    @post_generation
    def password(self, create, extracted, **kwargs):
        password = (
            extracted
            if extracted
            else Faker(
                "password",
                length=42,
                special_chars=True,
                digits=True,
                upper_case=True,
                lower_case=True,
            ).evaluate(None, None, extra={"locale": None})
        )
        self.set_password(password)

    @post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for group in extracted:
                self.groups.add(group)

    class Meta:
        model = get_user_model()
        django_get_or_create = ["username"]



"""An encapsulated thread-local variable that indicates whether future DB
writes should be "stuck" to the replica."""

import threading
from functools import wraps

_default_locals = threading.local()


def this_thread_is_pinned():
    """Return whether the current thread should send all its reads to the
    master DB."""
    return getattr(_default_locals, "pinned", False)


def pin_this_thread():
    """Mark this thread as "stuck" to the master for all DB access."""
    _default_locals.pinned = True


def unpin_this_thread():
    """Unmark this thread as "stuck" to the master for all DB access.
    If the thread wasn't marked, do nothing.
    """
    _default_locals.pinned = False


class UseDefaultDB:
    """A contextmanager/decorator to use the master database."""

    def __call__(self, func):
        @wraps(func)
        def decorator(*args, **kw):
            with self:
                return func(*args, **kw)

        return decorator

    def __enter__(self):
        _default_locals.old = this_thread_is_pinned()
        pin_this_thread()

    def __exit__(self, type, value, tb):
        if not _default_locals.old:
            unpin_this_thread()


_replica_locals = threading.local()


def this_replica_thread_is_pinned():
    """Return whether the current thread should send all its reads to the
    replica DB."""
    return getattr(_replica_locals, "pinned", False)


def pin_this_replica_thread():
    """Mark this thread as "stuck" to the replica for all DB access."""
    _replica_locals.pinned = True


def unpin_this_replica_thread():
    """Unmark this thread as "stuck" to the replica for all DB access.
    If the thread wasn't marked, do nothing.
    """
    _replica_locals.pinned = False


class UseReplicaDB:
    """A contextmanager/decorator to use the replica database."""

    def __call__(self, func):
        @wraps(func)
        def decorator(*args, **kw):
            with self:
                return func(*args, **kw)

        return decorator

    def __enter__(self):
        _replica_locals.old = this_replica_thread_is_pinned()
        pin_this_replica_thread()

    def __exit__(self, type, value, tb):
        if not _replica_locals.old:
            unpin_this_replica_thread()


use_default_db = UseDefaultDB()
use_replica_db = UseReplicaDB()



import os

from django.db import ProgrammingError, connection, migrations
from django.db.models import Q
from django_tenants.migration_executors import base, multiproc, standard
from django_tenants.utils import get_public_schema_name, get_tenant_model


class TenantFieldMixin:
    def __init__(self, *args, **kwargs):
        self.create_in_public_schema = kwargs.pop("public_schema", False)
        self.create_in_tenant_schema = kwargs.pop("tenant_schema", False)
        super().__init__(*args, **kwargs)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        if (self.create_in_public_schema and connection.schema_name == get_public_schema_name()) or (
            self.create_in_tenant_schema and connection.schema_name != get_public_schema_name()
        ):
            try:
                super().database_forwards(app_label, schema_editor, from_state, to_state)
            except ProgrammingError as excp:  # sometimes it can exist
                if "already exists" not in str(excp):
                    raise


class AddTenantField(TenantFieldMixin, migrations.AddField):
    pass


class RemoveTenantField(TenantFieldMixin, migrations.RemoveField):
    pass


class AddCustomField(TenantFieldMixin, migrations.AddField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.create_in_tenant_schema = False


def get_db_table():
    model = get_tenant_model()
    return model.objects._meta.db_table


class BaseSoftDeleteExecutorMixin:
    def __init__(self, args, options):
        super().__init__(args, options)
        self.inactive_tenants = set()
        self.multi_inactive_tenants = set()
        tenant_table = get_db_table()
        if tenant_table in connection.introspection.table_names():
            inactive_tenants = get_tenant_model().objects.filter(Q(is_deleted=True) | Q(is_ready=False))
            self.inactive_tenants = inactive_tenants.values_list("schema_name", flat=True).distinct()
            self.multi_inactive_tenants = inactive_tenants.values_list("schema_name", "tenant_type").distinct()

    def run_migrations(self, tenants=None):
        tenants = tenants or []
        active_tenants = list(set(tenants) - self.inactive_tenants)
        super().run_migrations(tenants=active_tenants)

    def run_multi_type_migrations(self, tenants=None):
        tenants = tenants or []
        active_tenants = list(set(tenants) - self.multi_inactive_tenants)
        super().run_multi_type_migrations(tenants=active_tenants)


class MultiprocessingSoftDeleteExecutor(BaseSoftDeleteExecutorMixin, multiproc.MultiprocessingExecutor):
    codename = "multiprocessing_soft_delete"


class StandardSoftDeleteExecutor(BaseSoftDeleteExecutorMixin, standard.StandardExecutor):
    codename = "standard_soft_delete"


def get_all_subclasses(cls):
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def get_executor(codename=None):
    codename = codename or os.environ.get("EXECUTOR", standard.StandardExecutor.codename)

    for klass in get_all_subclasses(base.MigrationExecutor):
        if klass.codename == codename:
            return klass

    raise NotImplementedError("No executor with codename %s" % codename)



import unittest

from ..converters import (
    clean_html,
    convert_numbers_to_words,
    parse_json_string,
    phone_format,
    snake_case,
)


class TestSnakeCase(unittest.TestCase):
    camel_case_name = "HelloWorld"
    snake_case_name = "hello_world"
    not_snake_case_name = "helloWorld"

    def test_snake_case_pass(self):
        assert snake_case(self.camel_case_name) == self.snake_case_name

    def test_snake_case_fail(self):
        assert snake_case(self.camel_case_name) != self.not_snake_case_name


class TestPhoneFormat(unittest.TestCase):
    phone_number_unformatted = "1234567890"
    phone_number_formatted = "123-456-7890"
    phone_number_int = 1234567890

    def test_phone_format_pass(self):
        assert phone_format(self.phone_number_unformatted) == self.phone_number_formatted

    def test_phone_number_integer_passed(self):
        with self.assertRaises(TypeError):
            phone_format(self.phone_number_int)


class TestCleanHtml(unittest.TestCase):
    hello_world_paragraph = "<p>Hello, World</p>"
    hello_world_script = '<script>alert("Hello World!");</script>'
    custom_sanitizer = {"tags": ["p"]}

    # Test with default sanitizer
    def test_clean_html_paragraph(self):
        assert clean_html(self.hello_world_paragraph) == self.hello_world_paragraph

    def test_clean_html_script(self):
        assert clean_html(self.hello_world_script) == ""

    # Test with invalid input
    def test_clean_html_invalid_input(self):
        assert clean_html("") == ""


class TestNumbersToWords(unittest.TestCase):
    def test_numbers_to_words_int_pass(self):
        assert convert_numbers_to_words(23) == "twenty-three"

    def test_numbers_to_words_float_pass(self):
        assert convert_numbers_to_words(23.3) == "twenty-three point three"

    def test_numbers_to_words_str_pass(self):
        assert convert_numbers_to_words("23") == "twenty-three"

    def test_numbers_to_words_float_as_a_string_pass(self):
        assert convert_numbers_to_words("23.3") == "twenty-three point three"

    def test_numbers_to_words_invalid_input(self):
        assert convert_numbers_to_words("Hello") == "zero"


class TestParseJsonString(unittest.TestCase):
    def test_parse_json_string_dict_pass(self):
        assert parse_json_string('{"Hello":"World"}') == {"Hello": "World"}

    def test_parse_json_string_non_dict_pass(self):
        assert parse_json_string("Hello") == "Hello"

    def test_parse_json_string_int_fail(self):
        with self.assertRaises(TypeError):
            parse_json_string(12)



from datetime import timedelta

import requests
from django.core.cache import cache
from django.utils import timezone
from requests.auth import HTTPBasicAuth
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    ProxyError,
    ReadTimeout,
    SSLError,
    Timeout,
    TooManyRedirects,
)
from rest_framework import status

# User properties
# https://docs.microsoft.com/en-us/graph/api/resources/user?view=graph-rest-1.0#properties


class MicrosoftGraph:
    access_token = None
    access_token_expiration = None

    def __init__(self, tenant_id, client_id, secret_key):
        self.tenant_id = tenant_id
        self.auth_url_base = f"https://login.microsoftonline.com/{self.tenant_id}/oauth2/v2.0"

        self.basic_auth = HTTPBasicAuth(client_id, secret_key)

        self.load_credentials()
        self.connection_exceptions = (
            ConnectionError,
            ProxyError,
            ReadTimeout,
            SSLError,
            Timeout,
            TooManyRedirects,
        )

    def get_manager(self, user_id, log_config):
        # user_id is either the UUID or userPrincipalName
        # https://docs.microsoft.com/en-us/graph/api/user-list-manager?view=graph-rest-1.0&tabs=http

        response = self.execute("get", f"users/{user_id}/manager", log_config)

        return response

    def get_users(self, log_config):
        # user_id is either the UUID or userPrincipalName
        # https://docs.microsoft.com/en-us/graph/api/user-list-manager?view=graph-rest-1.0&tabs=http
        # https://graph.microsoft.com/v1.0/users?$select=id,displayName,mail,jobTitle,mobilePhone,officeLocation,surname,givenName,department&$expand=manager($select=displayName,mail)
        response = self.execute(
            "get",
            "users/?$select=id,displayName,mail,jobTitle,mobilePhone,officeLocation,surname,givenName,department&$expand=manager($select=id,displayName,mail)",
            log_config,
        )

        return response

    def create_user(self, payload, log_config):
        # user_id is either the UUID or userPrincipalName
        # https://docs.microsoft.com/en-us/graph/api/user-list-manager?view=graph-rest-1.0&tabs=http

        response = self.execute("post", "users", log_config, json=payload)

        # if response.status_code == status.HTTP_201_CREATED:
        #     return response.json()

        return response

    def assign_manager(self, payload, user_id, log_config):
        return self.execute("put", f"users/{user_id}/manager/$ref", log_config, json=payload)

    def get_user_manager(self, user_principle_name=None, user_id=None, log_config=None, params=None):
        # GET /me/manager
        # GET /users/{id | userPrincipalName}/manager
        if user_principle_name is None and user_id is None:
            return None
        if user_principle_name:
            url = f"users/{user_principle_name}/manager"
        else:
            url = f"users/{user_id}/manager"

        response = self.execute("get", url, log_config, params=params)

        return response

    def get_user(self, user_principle_name=None, user_id=None, log_config=None, params=None):
        # user_id is either the UUID or userPrincipalName
        # https://docs.microsoft.com/en-us/graph/api/user-list-manager?view=graph-rest-1.0&tabs=http
        if user_principle_name is None and user_id is None:
            return None
        if user_principle_name:
            url = f"users/{user_principle_name}"
        else:
            url = f"users/{user_id}"

        response = self.execute("get", url, log_config, params=params)

        return response

    def get_all_groups(self, log_config=None):
        # user_id is either the UUID or userPrincipalName
        # https://docs.microsoft.com/en-us/graph/api/user-list-manager?view=graph-rest-1.0&tabs=http

        response = self.execute("get", "groups?$select=id,displayName", log_config)

        payload = response.json()
        del payload["@odata.context"]

        return payload

    def get_group_owners(self, group_id, log_config=None):
        # GET /groups/{id}/owners
        response = self.execute("get", f"groups/{group_id}/owners", log_config)

        return response

    def get_user_groups(self, user_principle_name=None, user_id=None, log_config=None):
        # GET https://graph.microsoft.com/v1.0/users/6e7b768e-07e2-4810-8459-485f84f8f204/memberOf
        response = None
        if user_id:
            response = self.execute("get", f"users/{user_id}/memberOf", log_config)
        elif user_principle_name:
            response = self.execute("get", f"users/{user_principle_name}/memberOf", log_config)
        else:
            raise KeyError("Mandatory field user_id or user_principle_name is unavailable")

        return response

    def get_group_members(self, group_id=None, log_config=None, params=None):
        # GET https://graph.microsoft.com/v1.0/users/6e7b768e-07e2-4810-8459-485f84f8f204/memberOf
        # query_params - ?$select=displayName, id, description
        url = f"groups/{group_id}/members"
        response = self.execute("get", url, log_config, params=params)

        return response

    def add_user_to_ad_group(self, user_id=None, group_id=None, log_config=None):
        # POST /groups/{group-id}/members/$ref
        payload = {"@odata.id": f"https://graph.microsoft.com/v1.0/directoryObjects/{user_id}"}
        response = self.execute("post", f"groups/{group_id}/members/$ref", log_config=log_config, json=payload)

        return response

    def remove_user_from_ad_group(self, user_id=None, group_id=None, log_config=None):
        response = self.execute("delete", f"groups/{group_id}/members/{user_id}/$ref", log_config=log_config)

        return response

    #
    # request utilities

    def execute(self, method, endpoint, log_config, **kwargs):
        method = method.lower()

        if not hasattr(requests, method):
            raise KeyError(f"Method is not available: {method}")

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        kwargs["headers"].update({"Authorization": f"Bearer {self.get_access_token()}"})

        url = f"https://graph.microsoft.com/v1.0/{endpoint}"
        request_time = timezone.now()
        response = requests.request(method=method, url=url, **kwargs)
        log_entry = None
        try:
            if log_config:
                if response:
                    response_time = request_time + response.elapsed
                else:
                    response_time = request_time

                log_entry = log_config["model"](
                    service_request=log_config["service_request"],
                    requested_by=log_config["user"],
                    request_url=f"{method}: {url}",
                    request_headers=kwargs["headers"],
                    response_headers=response.headers if response else "",
                    request_body=kwargs["json"] if method.lower() == "post" else "",
                    request_time=request_time,
                    response_code=response.status_code,
                    response_body="" if response.status_code == status.HTTP_204_NO_CONTENT else response.json(),
                    response_time=response_time,
                    active_directory_type="AZURE",
                )
                log_entry.save()

            response.raise_for_status()
            return response
        except self.connection_exceptions as excp:
            if log_config:
                if response:
                    response_time = log_entry.request_time + response.elapsed
                else:
                    response_time = log_entry.request_time
                log_entry = log_config["model"](
                    service_request=log_config["service_request"],
                    requested_by=log_config["user"],
                    request_url=f"{method}: {url}",
                    request_headers=kwargs["headers"],
                    response_headers="",
                    request_body=kwargs["json"] if method.lower() == "post" else "",
                    request_time=request_time,
                    response_code=0,
                    response_body=excp,
                    response_time=response_time,
                    active_directory_type="AZURE",
                )
                log_entry.save()
            raise
        except HTTPError as excp:
            raise Exception(
                f"[AZURE] - Received bad response. Status code: [{response.status_code}] for [{method}: {endpoint}]. Error: [{response.text}]"
            ) from excp

    #
    # auth utilities

    def load_credentials(self):
        credentials = cache.get("ms_azure_backend_access_token")
        if credentials is None:
            return
        self.access_token = credentials.get("access_token")
        self.access_token_expiration = credentials.get("expiration")

    def save_credentials(self):
        credentials = {
            "access_token": self.access_token,
            "expiration": self.access_token_expiration,
        }
        cache.set(
            "ms_azure_backend_access_token",
            credentials,
            timeout=(self.access_token_expiration - timezone.now()).seconds,
        )

    def is_token_expired(self):
        if not self.access_token_expiration:
            return True

        return timezone.now() > self.access_token_expiration

    def get_access_token(self):
        if self.access_token and not self.is_token_expired():
            return self.access_token

        url = f"{self.auth_url_base}/token"
        response = requests.post(
            url=url,
            auth=self.basic_auth,
            data={"grant_type": "client_credentials", "scope": "https://graph.microsoft.com/.default"},
        )

        payload = response.json()

        if response.status_code in (400, 401, 404, 415):
            raise Exception(
                f"[AZURE] - Received bad response. Status code: [{response.status_code}] for [post: {url}]. Error: [{payload['error']}: {payload['error_description']}]"
            )

        self.access_token = payload.get("access_token")
        self.access_token_expiration = timezone.now() + timedelta(seconds=payload.get("expires_in"))

        self.save_credentials()

        return self.access_token



import logging

import sentry_sdk
from django.conf import settings
from smartystreets_python_sdk import ClientBuilder, StaticCredentials, exceptions
from smartystreets_python_sdk.us_autocomplete import Lookup as AutocompleteLookup
from smartystreets_python_sdk.us_extract import Lookup as ExtractLookup
from smartystreets_python_sdk.us_street import Lookup as StreetLookup

LOGGER = logging.getLogger("root")


class SmartyStreets:
    def __init__(self):
        if not settings.SMARTY_STREETS_AUTH_ID or not settings.SMARTY_STREETS_AUTH_TOKEN:
            e = Exception("Smarty Street Credentials NOT set")
            sentry_sdk.capture_exception(e)
            return e

        credentials = StaticCredentials(settings.SMARTY_STREETS_AUTH_ID, settings.SMARTY_STREETS_AUTH_TOKEN)
        self.client = ClientBuilder(credentials).build_us_street_api_client()
        self.auto_complete_client = ClientBuilder(credentials).build_us_autocomplete_api_client()
        self.extract_client = ClientBuilder(credentials).build_us_extract_api_client()

    # Documentation for input fields can be found at:
    # https://smartystreets.com/docs/us-street-api#input-fields
    def address_lookup(
        self,
        addressee=None,
        street=None,
        street2=None,
        secondary=None,
        urbanization=None,
        city=None,
        state=None,
        zipcode=None,
        candidates=5,
        match="invalid",
    ):
        if not street or not state:
            return

        lookup = StreetLookup()
        # lookup.input_id = "24601"  # Optional ID from your system
        lookup.addressee = addressee
        lookup.street = street
        lookup.street2 = street2
        lookup.secondary = secondary
        lookup.urbanization = urbanization  # Only applies to Puerto Rico addresses
        lookup.city = city
        lookup.state = state
        lookup.zipcode = zipcode
        lookup.candidates = candidates
        lookup.match = match  # "invalid" is the most permissive match,
        # this will always return at least one result even if the address is invalid.
        # Refer to the documentation for additional Match Strategy options.

        try:
            self.client.send_lookup(lookup)
        except exceptions.SmartyException as err:
            sentry_sdk.capture_exception(err)

        addresses = []
        for address in lookup.result:
            temp_address = address.__dict__
            temp_address["components"] = address.components.__dict__
            temp_address["metadata"] = address.metadata.__dict__
            temp_address["analysis"] = address.analysis.__dict__
            try:
                del temp_address["analysis"]["is_ews_match"]
            except Exception as excp:
                LOGGER.exception(f"Error: [{excp}]")
            addresses.append(temp_address)

        return addresses

    # Documentation for input fields can be found at:
    # https://smartystreets.com/docs/us-autocomplete-api#http-request-input-fields
    def auto_complete_address_lookup(self, address_string, city_filter=None, state_filter=None, max_suggestions=5):
        lookup = AutocompleteLookup(address_string)
        if city_filter:
            lookup.add_city_filter = city_filter
        if state_filter:
            lookup.add_state_filter = state_filter
        lookup.max_suggestions = max_suggestions

        try:
            self.auto_complete_client.send(lookup)
        except exceptions.SmartyException as err:
            sentry_sdk.capture_exception(err)

        addresses = []
        for suggestion in lookup.result:
            addresses.append(suggestion.__dict__)
        return addresses

    # Documentation for input fields can be found at:
    # https://smartystreets.com/docs/cloud/us-extract-api#http-request-input-fields
    def extract_address(
        self,
        address_string,
        aggressive=True,
        addresses_have_line_breaks=False,
        addresses_per_line=1,
    ):
        lookup = ExtractLookup()
        lookup.text = address_string
        lookup.aggressive = aggressive
        lookup.addresses_have_line_breaks = addresses_have_line_breaks
        lookup.addresses_per_line = addresses_per_line

        try:
            addresses = self.extract_client.send(lookup).addresses
        except exceptions.SmartyException as err:
            sentry_sdk.capture_exception(err)

        for address in addresses:
            if address.verified and len(address.candidates) == 1:
                return address.candidates[0].components.__dict__

        return None



from constance import config
from django_tenants.utils import (
    get_public_schema_name,
    get_tenant_model,
    schema_context,
)
from django_twilio_2fa.utils import get_setting, get_twilio_client

__all__ = ["create_twilio_verify_service"]


def create_twilio_verify_service(schema_name, friendly_name=None, force=False, use_existing=True):
    tenant = None

    if schema_name != get_public_schema_name():
        tenant = get_tenant_model().objects.get(schema_name=schema_name)

    # Twilio credential check
    if not get_setting("ACCOUNT_SID"):
        return (
            False,
            "Twilio credentials are not available. Set TWILIO_2FA_ACCOUNT_SID and TWILIO_2FA_AUTH_TOKEN settings.",
        )

    # Set friendly name
    if not friendly_name and tenant:
        friendly_name = str(tenant)
    elif not friendly_name:
        friendly_name = "LOS"

    with schema_context(schema_name):
        # Check for existing config
        has_verify_sid = config.TWILIO_2FA_SERVICE_ID != ""

        if has_verify_sid and not force:
            return (False, f"Twilio Verify service ID is already set for {schema_name}")

        twilio = get_twilio_client()

        # Check for existing service
        existing_service_sid = None

        if use_existing:
            print(f"Searching for Twilio Verify service with friendly name: {friendly_name}")
            services = twilio.verify.v2.services.list(limit=100, page_size=100)

            for service in services:
                if service.friendly_name != friendly_name:
                    continue

                existing_service_sid = service.sid
                break

        if existing_service_sid:
            config.TWILIO_2FA_SERVICE_ID = existing_service_sid

            return (True, f"Set Twilio Verify service for {schema_name} with existing SID: {existing_service_sid}")

        service = twilio.verify.v2.services.create(friendly_name=friendly_name)

        config.TWILIO_2FA_SERVICE_ID = service.sid

        return (True, f"Created new Twilio Verify service for {schema_name}: {config.TWILIO_2FA_SERVICE_ID}")



from constance import config
from django import template
from django.conf import settings
from django.templatetags.static import static
from django.utils.safestring import mark_safe

register = template.Library()

v3_site_key = getattr(settings, "RECAPTCHA_V3_SITE_KEY", "")
v3_secret_key = getattr(settings, "RECAPTCHA_V3_SECRET_KEY", "")
v2_site_key = getattr(settings, "RECAPTCHA_V2_SITE_KEY", "")
v2_secret_key = getattr(settings, "RECAPTCHA_V2_SECRET_KEY", "")
vue_recaptcha_url = static("vendor/vue-recaptcha/vue-recaptcha.min.js")


@register.simple_tag
def RECAPTCHA_UPPER_SCORE():
    return config.GOOGLE_RECAPTCHA_V3_UPPER_SCORE


@register.simple_tag
def RECAPTCHA_LOWER_SCORE():
    return config.GOOGLE_RECAPTCHA_V3_LOWER_SCORE


@register.simple_tag
def RECAPTCHA_V3_SITE_KEY():
    return v3_site_key


@register.simple_tag
def RECAPTCHA_V3_SECRET_KEY():
    return v3_secret_key


@register.simple_tag
def RECAPTCHA_V2_SITE_KEY():
    return v2_site_key


@register.simple_tag
def RECAPTCHA_V2_SECRET_KEY():
    return v2_secret_key


@register.simple_tag
def RECAPTCHA_SCRIPTS_INCLUDE():
    html = f'<script src="https://www.google.com/recaptcha/api.js?render={v3_site_key}"></script>'
    html += '<script src="https://www.google.com/recaptcha/api.js?onload=vueRecaptchaApiLoaded&render=explicit" async defer></script>'
    html += f'<script src="{vue_recaptcha_url}"></script>'
    html += "<style>.grecaptcha-badge {{ opacity:0;}}</style>"
    return mark_safe(html)  # nosemgrep


@register.simple_tag
def RECAPTCHA_SCRIPT():
    html = (
        f"grecaptcha.ready(function () {{"
        f"grecaptcha.execute('{v3_site_key}', {{ action: 'load' }}).then(function (token) {{"
        f'axios.post("/api/services/google/re_captcha",'
        f"                  {{response: token}}"
        f"                  ).then(function (response) {{"
        f"            let score= response.data.data.score;"
        f"            vm.score= score;"
        f"            if(score > {RECAPTCHA_UPPER_SCORE()}){{"
        f"              vm.showv2= false;"
        f"            }}else if({RECAPTCHA_UPPER_SCORE()} >= score > {RECAPTCHA_LOWER_SCORE()}){{"
        f"              vm.showv2= true;"
        f"            }}else{{"
        f"              vm.showv2= true;"
        f"            }}"
        f"          }}).catch(function (error) {{"
        f"              vm.showv2= true;"
        f"          }});"
        f"        }});"
        f"      }});   "
    )

    return mark_safe(html)  # nosemgrep



import io

from ..io_utils import BytesIOWrapper


class TestBytesIOWrapper:
    def test_bytes_io_wrapper_read(self):
        text_io_buffer = io.StringIO("Hello World!")
        bytes_io = BytesIOWrapper(text_io_buffer)
        assert bytes_io.read() == b"Hello World!"

    # TODO: Fix the test below.
    # def test_bytes_io_wrapper_read1(self):
    #     text_io_buffer = io.StringIO("Hello World!")
    #     bytes_io = BytesIOWrapper(text_io_buffer)
    #     result = bytes_io.read1(5)
    #     assert result.decode() == "Hello"

    def test_bytes_io_wrapper_peek(self):
        text_io_buffer = io.StringIO("Hello World!")
        bytes_io = BytesIOWrapper(text_io_buffer)
        assert bytes_io.read(1) == b"H"
        # Reset the stream position to the beginning
        text_io_buffer.seek(0)
        assert bytes_io.read(1) == b"H"

    def test_bytes_io_wrapper_with_encoding(self):
        text_io_buffer = io.StringIO("Hello World!")
        bytes_io = BytesIOWrapper(text_io_buffer, encoding="utf-16")
        assert bytes_io.read() == b"\xff\xfeH\x00e\x00l\x00l\x00o\x00 \x00W\x00o\x00r\x00l\x00d\x00!\x00"

    def test_bytes_io_wrapper_with_errors(self):
        text_io_buffer = io.StringIO("Hello World!")
        bytes_io = BytesIOWrapper(text_io_buffer, encoding="ascii", errors="replace")
        assert bytes_io.read() == b"Hello World!"



def superusers_and_external_services(*, hijacker=None, hijacked=None):
    """
    Superusers, global superusers, and external service team users may hijack other users.
    """
    if hijacker.is_superuser or hijacker.is_global_superuser:
        return True

    if hijacker.is_external_service_team_user:
        return True



from django.conf import settings
from django.db import models
from django.utils import timezone


class SoftDeleteManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(is_deleted=False)


class SoftDeletableModel(models.Model):
    is_deleted = models.BooleanField(default=False)
    deleted_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        editable=False,
        on_delete=models.PROTECT,
    )
    deleted_at = models.DateTimeField(blank=True, null=True)

    objects = models.Manager()
    available_objects = SoftDeleteManager()

    def delete(self, actor=None, soft=True, *args, **kwargs):
        if soft:
            deleted_date = timezone.now()
            self.__class__.objects.filter(pk=self.pk).update(is_deleted=True, deleted_at=deleted_date, deleted_by=actor)
        else:
            super().delete(*args, **kwargs)

    class Meta:
        abstract = True



import logging

from django.contrib import admin
from django_tenants.admin import TenantAdminMixin
from django_tenants.utils import get_public_schema_name

from .paginators import FastPaginator

IS_POPUP_VAR = "_popup"
TO_FIELD_VAR = "_to_field"

LOGGER = logging.getLogger("root")


class AutoCompleteAdminMixin(admin.ModelAdmin):
    AUTOCOMPLETE_FIELD_TYPES = [
        "ForeignKey",
        "ManyToManyField",
    ]

    def get_autocomplete_field_types(self):
        return self.AUTOCOMPLETE_FIELD_TYPES

    def get_autocomplete_fields(self, request):
        autocomplete_fields = list(super().get_autocomplete_fields(request))

        for field in self.model._meta.get_fields():
            if type(field).__name__ not in self.get_autocomplete_field_types():
                continue

            try:
                admin_class = admin.site._registry[field.related_model]
                if not hasattr(admin_class, "search_fields") or not admin_class.search_fields:
                    # Skip if no search_fields present (throws an exception)
                    continue
            except Exception as excp:
                # Just Logging the exception
                LOGGER.debug(f"Error: [{excp}]")

            autocomplete_fields.append(field.name)

        return autocomplete_fields


class FastPaginationModelAdmin(AutoCompleteAdminMixin):
    paginator = FastPaginator
    show_full_result_count = False


class PublicTenantOnlyAdmin(TenantAdminMixin, FastPaginationModelAdmin):
    """
    Hides public models from tenants.
    django_tenants.middleware.main.TenantMainMiddleware adds a tenant attribute if a schema is found but does not add if it is the public schema
    """

    def has_view_permission(self, request, *args, view=None, **kwargs):
        return not hasattr(request, "tenant") or request.tenant.schema_name == get_public_schema_name()

    def has_add_permission(self, request, *args, view=None, **kwargs):
        return not hasattr(request, "tenant") or request.tenant.schema_name == get_public_schema_name()

    def has_change_permission(self, request, *args, view=None, **kwargs):
        return not hasattr(request, "tenant") or request.tenant.schema_name == get_public_schema_name()

    def has_delete_permission(self, request, *args, view=None, **kwargs):
        return not hasattr(request, "tenant") or request.tenant.schema_name == get_public_schema_name()

    def has_view_or_change_permission(self, request, *args, view=None, **kwargs):
        return not hasattr(request, "tenant") or request.tenant.schema_name == get_public_schema_name()



from django.db import models
from django.utils.translation import gettext_lazy as _


class PERMISSION_TYPE(models.IntegerChoices):
    """List of permission types for auth permissions."""

    DJANGO_DEFAULT = 1, _("Django Default")
    BACK_OFFICE = 2, _("Back Office")
    ADMINISTRATIVE = 3, _("Administrative")
    OTHER = 4, _("Other")


class GROUP_TYPE(models.IntegerChoices):
    """List of group types for auth groups."""

    BACK_OFFICE = 1, _("Back Office")
    ADMINISTRATIVE = 2, _("Administrative")
    OTHER = 3, _("Other")
    SUPERVISOR = 4, _("Supervisor")
    DEFAULT = 5, _("Default")



from django.conf import settings
from django.db import connection
from django.urls import resolve, reverse
from django.utils.functional import lazy
from django_tenants.utils import get_public_schema_name


def get_urlconf():
    if connection.schema_name == get_public_schema_name():
        return settings.TENANT_TYPES["public"]["URLCONF"]
    elif connection.schema_name == "filemover":
        return settings.TENANT_TYPES["filemover"]["URLCONF"]
    else:
        return settings.TENANT_TYPES["organization"]["URLCONF"]


def custom_reverse(viewname, url_conf=None, args=None, kwargs=None, current_app=None):
    """
    Custom reverse function for tenant-aware url resolving
    """
    url_conf = get_urlconf() if url_conf is None else url_conf

    return reverse(  # nosemgrep
        viewname,
        urlconf=url_conf,
        args=args,
        kwargs=kwargs,
        current_app=current_app,
    )


custom_reverse_lazy = lazy(custom_reverse, str)


def custom_resolve(path, urlconf=None):
    """
    Custom function for tenant-aware url resolving
    """
    if urlconf is None:
        urlconf = get_urlconf()
    return resolve(path, urlconf=urlconf)



from datetime import timedelta

from django.conf import settings
from django.contrib.staticfiles.storage import ManifestFilesMixin
from django.utils import timezone
from django.utils.encoding import filepath_to_uri
from storages.backends.s3boto3 import S3Boto3Storage


class StaticRootS3Boto3Storage(ManifestFilesMixin, S3Boto3Storage):
    location = "static"
    bucket_name = getattr(settings, "AWS_STATIC_STORAGE_BUCKET_NAME", None)  # local env breaks w/o this logic
    default_acl = "private"


class MediaRootS3Boto3Storage(S3Boto3Storage):
    location = "media"
    file_overwrite = False
    default_acl = "private"
    custom_domain = False

    def url(self, name, parameters=None, expire=None, http_method=None):
        """
        Replace internal domain with custom domain for signed URLs.
        See issue: https://github.com/jschneier/django-storages/issues/165#issuecomment-810166563
        """
        try:
            # put this in a try block for now, so I don't break everything
            from constance import config

            constance_expire = config.AWS_QUERYSTRING_EXPIRE
            url = super().url(name, parameters, constance_expire, http_method)
        except Exception:
            url = super().url(name, parameters, expire, http_method)
        custom_url = url.replace(
            settings.AWS_S3_CUSTOM_DOMAIN,
            f"{settings.AWS_S3_URL_PROTOCOL}//{settings.AWS_S3_CUSTOM_MEDIA_DOMAIN}",
        )
        return custom_url


class PublicMediaRootS3Boto3Storage(S3Boto3Storage):
    bucket_name = "brandings"
    custom_domain = settings.AWS_S3_CUSTOM_DOMAIN
    endpoint_url = f"https://{settings.AWS_S3_CUSTOM_DOMAIN}"
    default_acl = "private"
    file_overwrite = False
    querystring_auth = False

    def url(self, name, parameters=None, expire=None, http_method=None):
        name = self._normalize_name(self._clean_name(name))
        if expire is None:
            expire = self.querystring_expire

        if self.custom_domain:
            url = "{}/{}/{}".format(settings.AWS_PUBLIC_CLOUDFRONT_DOMAIN, self.bucket_name, filepath_to_uri(name))

            if self.querystring_auth and self.cloudfront_signer:
                expiration = timezone.now() + timedelta(seconds=expire)

                return self.cloudfront_signer.generate_presigned_url(url, date_less_than=expiration)

            return url

        params = parameters.copy() if parameters else {}
        params["Bucket"] = self.bucket.name
        params["Key"] = name
        url = self.bucket.meta.client.generate_presigned_url(
            "get_object", Params=params, ExpiresIn=expire, HttpMethod=http_method
        )
        if self.querystring_auth:
            return url
        return self._strip_signing_parameters(url)


class LocalPublicMediaRootS3Boto3Storage(PublicMediaRootS3Boto3Storage):
    endpoint_url = f"http://{settings.AWS_S3_CUSTOM_DOMAIN}"



import io


class BytesIOWrapper(io.BufferedReader):
    """Wrap a buffered bytes stream over TextIOBase string stream."""

    def __init__(self, text_io_buffer, encoding=None, errors=None, **kwargs):
        super(BytesIOWrapper, self).__init__(text_io_buffer, **kwargs)
        self.encoding = encoding or text_io_buffer.encoding or "utf-8"
        self.errors = errors or text_io_buffer.errors or "strict"

    def _encoding_call(self, method_name, *args, **kwargs):
        raw_method = getattr(self.raw, method_name)
        val = raw_method(*args, **kwargs)
        return val.encode(self.encoding, errors=self.errors)

    def read(self, size=-1):
        return self._encoding_call("read", size)

    def read1(self, size=-1):
        return self._encoding_call("read1", size)

    def peek(self, size=-1):
        return self._encoding_call("peek", size)



import os
import re

import magic
import phonenumber_field.phonenumber
import phonenumbers
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.translation import gettext_lazy as _

from django_lc_utils.errors.error_manager import error

from .config.exceptions import fmt_error_msg

NUMERIC_PATTERN = r"[^0-9\s]"
alphanumeric_pattern = r"[^a-zA-Z0-9\s]"
numeric_pattern = r"[^0-9\s]"
numeric = RegexValidator(r"^\d*$", "Numeric Values Required")
decimal = RegexValidator(r"^\d*[.]?\d*$", "Decimal Values Required")
alphanumeric = RegexValidator(r"^[a-zA-Z0-9]*$", "Must Be Alphanumeric", code="alphanumeric")


VALID_KINDS = {
    "json": [
        "ASCII text",
        "JSON data",
    ],
    "jpg": [
        "JPEG image data",
    ],
    "jpeg": [
        "JPEG image data",
    ],
    "png": [
        "PNG image data",
    ],
    "csv": [
        "ASCII text",
    ],
    "pdf": [
        "PDF document",
    ],
    "xlsx": [
        "Microsoft Excel 2007+",
    ],
    "docx": [
        "Microsoft Word 2007+",
    ],
    # glitch w < Office 2007 docs .. xls and doc will not report correctly.
    # reports "Composite Document File V2 Document" instead
    "xls": [
        "Microsoft Excel 2003",
    ],
    "doc": [
        "Microsoft Word 2003",
    ],
    "pfx": [
        "data",
    ],
}


def validate_excel_file_extension(value):
    ext = os.path.splitext(value.name)[1]  # [0] returns path+filename
    valid_extensions = [".xls", ".xlsx", ".csv"]
    if ext.lower() not in valid_extensions:
        raise ValidationError(error.error_message("E0001"))


def validation_logic(value, magic_buffer_size, valid_kinds):
    ext_from_filename = os.path.splitext(value.name)[1].lower()[1:]
    ext_from_filename = "jpg" if ext_from_filename == "jpeg" else ext_from_filename  # conform jpg/jpeg
    expected_kind = valid_kinds.get(ext_from_filename)
    kind = magic.from_buffer(value.read(magic_buffer_size)).split(",")[0]
    if (expected_kind is None) or (kind not in expected_kind and kind != "Composite Document File V2 Document"):
        raise ValidationError(error.error_message("E0002"))
    elif kind == "Composite Document File V2 Document":
        # special case -  glitch w < Office 2007 docs.  validate further
        special_magic = bytes([0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1])
        # same string for all office docs so ppt pub or visio could slip through,  cannot differentiate between doc/xls
        value.seek(0)
        if value.read(len(special_magic)) != special_magic:
            raise ValidationError(error.error_message("E0003"))


def validate_file_extension(value):
    """
    Determine if file has appropriate extension and type.  Refactord to use python-magic module to check buffered
    contents rather than trust extension or trust header mime type.  For Office docs < Office 2007, further validation
    done with magic numbeers string due to reported type glitch
    https://pypi.org/project/python-magic/
    """
    magic_buffer_size = 2048
    valid_kinds = VALID_KINDS.copy()

    validation_logic(value, magic_buffer_size, valid_kinds)


def validate_file_extension_zip(value):
    """
    validate_file_extension function with the addition of zip file to VALID_KINDS
    """
    magic_buffer_size = 2048
    valid_kinds = VALID_KINDS.copy()
    valid_kinds.update(
        {
            "zip": [
                "Zip archive data",
            ]
        }
    )
    validation_logic(value, magic_buffer_size, valid_kinds)


def values_within_tolerance(val1, val2, tolerance=None, tolerance_pct=None):
    """
    Determine if val1 and val2 are the same or within the provided tolerances.
    If tolerance and tolerance_pct aren't provided, no tolerance will be allowed
    """
    val1 = float(val1 or 0)
    val2 = float(val2 or 0)
    tolerance = float(tolerance or 0)
    tolerance_pct = float(tolerance_pct or 0)

    if val1 and val1 > 0:
        variance1 = abs(round(float(val2) - float(val1), 2)) / float(val1)
    else:
        variance1 = 1

    variance2 = abs(float(val1) - float(val2))

    if variance1 <= tolerance_pct or round(variance2, 2) <= tolerance:
        return True

    return False


def validate_no_html_tags(value):
    html_matches = re.findall("^(?:.*<[^>]+>).*", value)
    if len(html_matches):
        err_code = "E1100"
        usr_msg = _("No HTML tags allowed in field.")
        tech_msg = "No HTML tags allowed in field."
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)


def validate_no_nested_html(value):
    if not value:
        return

    if isinstance(value, (list, tuple, set)):
        for nested_value in value:
            validate_no_nested_html(nested_value)
    elif isinstance(value, dict):
        for key, nested_value in value.items():
            validate_no_html_tags(str(key))
            validate_no_nested_html(nested_value)
    else:
        validate_no_html_tags(str(value))


def validate_phone_number(phone_number):
    phone_number = phone_number.replace("(", "").replace(")", "").replace("-", "").replace(" ", "")
    if phone_number and not len(phone_number) == 10:
        err_code = "E1101"
        usr_msg = _("Phone number must be 10 digits")
        tech_msg = "Phone number must be 10 digits"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)
    if re.findall(NUMERIC_PATTERN, phone_number):
        err_code = "E1102"
        usr_msg = _("Phone Number is Invalid")
        tech_msg = "Phone Number is Invalid"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)

    try:
        if "+1" not in phone_number:
            phone_number = f"+1 {phone_number}"
        phonenumbers.parse(phone_number)
    except Exception:
        err_code = "E1103"
        usr_msg = _("Phone Number is Invalid")
        tech_msg = "Phone Number is Invalid"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)


def validate_international_phone_number(phone_number):
    if not phone_number:
        return
    try:
        phonenumber_field.phonenumber.PhoneNumber.from_string(phone_number)
    except Exception:
        err_code = "E1103"
        usr_msg = _("Phone Number is Invalid")
        tech_msg = "Phone Number is Invalid"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)


def validate_max_80_length(field):
    if (len(field) + (len(re.findall(alphanumeric_pattern, field)))) > 80:
        err_code = "E1104"
        usr_msg = _("Must be less than 80 characters (non-alphanumeric characters will count as 2 characters)")
        tech_msg = "Must be less than 80 characters (non-alphanumeric characters will count as 2 characters)"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)


def validate_domain(value):
    regex = re.compile(
        # r"^(?:http|ftp)s?://" # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
        # domain...
        r"localhost|" r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # localhost...
        # ...or ip
        # r"(?::\d+)?" # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    if re.match(regex, str(value)) is None:
        err_code = "E1105"
        usr_msg = _("Invalid domain provided")
        tech_msg = "Invalid domain provided"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)

    invalid_schemes = ["http://", "https://", "www."]
    if any(scheme in value for scheme in invalid_schemes):
        err_code = "E1106"
        usr_msg = _("Do not include any of the following in the domain: {}").format(invalid_schemes)
        tech_msg = f"Do not include any of the following in the domain: {invalid_schemes}"
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)

    return value


def alphanumeric_with_underscore(value):
    reg = re.compile("^[a-zA-Z0-9_]*$")
    if not reg.match(value):
        raise ValidationError("Only Alphanumeric/underscore(_) is allowed")


def validate_cif_file_extension(value):
    """
    Determine if file has appropriate extension and type.  Refactord to use python-magic module to check buffered
    contents rather than trust extension or trust header mime type.  For Office docs < Office 2007, further validation
    done with magic numbeers string due to reported type glitch
    https://pypi.org/project/python-magic/
    """
    magic_buffer_size = 2048
    valid_kinds = {
        "txt": "ASCII text",
        "csv": "ASCII text",
    }

    ext_from_filename = os.path.splitext(value.name)[1].lower()[1:]
    expected_kind = valid_kinds.get(ext_from_filename)
    kind = magic.from_buffer(value.read(magic_buffer_size)).split(",")[0]
    if kind != expected_kind:
        err_code = "E1501"
        usr_msg = _("Unsupported/invalid file extension.")
        tech_msg = "Unsupported/invalid file extension."
        error_msg = fmt_error_msg(err_code=err_code, usr_msg=usr_msg, tech_msg=tech_msg)
        raise ValidationError(error_msg)



from functools import wraps

from django.apps import apps
from django.contrib.auth.management import create_permissions
from django.shortcuts import redirect
from django_tenants.utils import get_public_schema_name


def lender_view(view_func):
    """
    Decorator for making sure lender routes are only accessed through Lender Portal
    """

    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        if not request.user.is_authenticated or request.user.is_borrower:
            return redirect("/")
        return view_func(request, *args, **kwargs)

    return _wrapped_view


def borrower_view(view_func):
    """
    Decorator for making sure borrower routes are only accessed through Borrower Portal
    """

    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        if not request.user.is_authenticated or not request.user.is_borrower:
            return redirect("/")
        return view_func(request, *args, **kwargs)

    return _wrapped_view


def tenant_migration(*args, tenant_schema=True, public_schema=False):
    """
    Decorator to control which schemas a data migration will execute under.
    If `tenant_schema=True`, the data migration will execute on non-public schemas.
    If `public_schema=True`, the data migration will execute on the public schema.
    """

    def _tenant_migration(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                _, schema_editor = args  # noqa
            except Exception as excp:
                raise Exception(f"Decorator requires apps & schema_editor as positional arguments: [{excp}]")

            result = None
            if (tenant_schema is True and schema_editor.connection.schema_name != get_public_schema_name()) or (
                public_schema is True and schema_editor.connection.schema_name == get_public_schema_name()
            ):
                result = func(*args, **kwargs)

            return result  # return value for testing purposes

        return wrapper

    if len(args) == 1 and callable(args[0]):
        return _tenant_migration(args[0])

    return _tenant_migration


def migrate_permissions(app_label=None):
    """
    Ensures the permissions for a given app (or all apps) are migrated before running a data migration.
    Args:
        app_label : Label of the app to force permissions to migrate; if none, performs all apps. Defaults to None.
    """

    def _migrate_permissions(func):
        def wrapper(*args, **kwargs):
            configs = apps.get_app_configs() if app_label is None else [apps.get_app_config(app_label)]

            for app_config in configs:
                app_config.models_module = True
                create_permissions(app_config, verbosity=0)
                app_config.models_module = None

            func(*args, **kwargs)

        return wrapper

    return _migrate_permissions



from django.contrib.admin.views.main import ChangeList
from django.db import connections


class RawChangeList(ChangeList):
    """
    Extended Django ChangeList to be able show data from RawQueryset.
    """

    def get_count(self):
        connection = connections[self.queryset.db]
        with connection.cursor() as c:
            if connection.vendor == "microsoft":  # CTE in subquery is not working in SQL Server
                c.execute(self.queryset.raw_query)
                c.execute("SELECT @@ROWCOUNT")
            else:
                query = "SELECT COUNT(*) FROM ({query}) AS sq"
                c.execute(query.format(query=self.queryset.raw_query))

            return c.fetchone()[0]

    def get_queryset_slice(self):
        connection = connections[self.queryset.db]
        if connection.vendor == "microsoft":
            # SQL Server needs ordered query for slicing
            if hasattr(self.queryset, "ordered") and self.queryset.ordered:
                query = "{query}"
            else:
                query = "{query} ORDER BY 1"
            query += " OFFSET {offset} ROWS FETCH NEXT {limit} ROWS ONLY"
        else:
            query = "{query} LIMIT {limit} OFFSET {offset}"

        # nosemgrep
        return self.queryset.model.objects.raw(
            query.format(
                query=self.queryset.raw_query,
                offset=self.page_num * self.list_per_page,
                limit=(self.page_num + 1) * self.list_per_page - self.page_num * self.list_per_page,
            )
        )

    def get_queryset(self, request):
        """
        Overriding to avoid applying filters in ChangeList because RawQueryset has not filter method.
        So any filters has to be applied manually for now.
        """
        qs = self.root_queryset
        if not hasattr(qs, "count"):
            qs.count = lambda: self.get_count()
        return qs

    def get_results(self, request):
        if self.show_all:
            qs = self.queryset
        else:
            qs = self.get_queryset_slice()

        paginator = self.model_admin.get_paginator(request, self.queryset, self.list_per_page)

        self.result_count = paginator.count
        self.show_full_result_count = False
        self.show_admin_actions = True
        self.full_result_count = 0
        self.result_list = list(qs)
        self.can_show_all = True
        self.multi_page = True
        self.paginator = paginator



from django.core.validators import MinLengthValidator
from django.db.models import CharField
from django.utils.translation import gettext_lazy as _
from phonenumber_field.modelfields import PhoneNumberField as BasePhoneNumberField

from .validators import numeric, validate_international_phone_number


class PhoneNumberField(BasePhoneNumberField):
    default_validators = [validate_international_phone_number]


class Rot26EnrcyptionField(CharField):
    def to_python(self, value) -> str | None:
        """Convert a string to a ROT26 encrypted string

        Returns:
            A ROT26 encrypted string
        """
        enc_constant = 26

        if value is None:
            return value

        rot26 = ""
        for v in value:
            o = ord(v)
            for x, y in enumerate(range(0, enc_constant)[::-1]):
                o = ((max((o ^ enc_constant) - (o | enc_constant), 0)) + o) ** (enc_constant - (x + y))
            rot26 += chr(o)
        return rot26


class TinField(CharField):
    LENGTH = 9

    default_validators = [MinLengthValidator(LENGTH, message=_(f"TIN must be {LENGTH} digits")), numeric]
    description = _("TIN")

    def __init__(self, **kwargs):
        kwargs.setdefault("max_length", self.LENGTH)
        kwargs.setdefault("db_index", True)

        super().__init__(**kwargs)


class ZipCodeField(CharField):
    LENGTH = 5

    description = _("Zip Code")
    default_validators = [MinLengthValidator(LENGTH, message=_(f"{description} must be {LENGTH} digits")), numeric]

    def __init__(self, **kwargs):
        kwargs.setdefault("max_length", self.LENGTH)
        kwargs.setdefault("null", True)
        kwargs.setdefault("blank", True)

        super().__init__(**kwargs)


class ZipCodePlus4Field(ZipCodeField):
    LENGTH = 4

    description = _("Zip+4 Code")
    default_validators = [MinLengthValidator(LENGTH, message=_(f"{description} must be {LENGTH} digits")), numeric]



from rest_framework import status, viewsets
from rest_framework.response import Response

from .mixins import ReadReplicaMixin


# FIXME: Add and fix readreplicamixin
class BorrowerPermissionModelViewSet(ReadReplicaMixin, viewsets.ModelViewSet):
    borrower_http_methods = ["get", "post"]

    def is_valid_user(self, request):
        if request.user.is_authenticated is False or (
            request.user.is_borrower
            and request.method.lower() not in map(lambda x: x.lower(), self.borrower_http_methods)
        ):
            return False

        return True

    def create(self, request, *args, **kwargs):
        if self.is_valid_user(request) is False:
            return Response(status=status.HTTP_404_NOT_FOUND)

        return super().create(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        if self.is_valid_user(request) is False:
            return Response(status=status.HTTP_404_NOT_FOUND)

        return super().partial_update(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        if self.is_valid_user(request) is False:
            return Response(status=status.HTTP_404_NOT_FOUND)

        return super().update(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        if self.is_valid_user(request) is False:
            return Response(status=status.HTTP_404_NOT_FOUND)

        return super().list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        if self.is_valid_user(request) is False:
            return Response(status=status.HTTP_404_NOT_FOUND)

        return super().retrieve(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        if self.is_valid_user(request) is False:
            return Response(status=status.HTTP_404_NOT_FOUND)

        return super().destroy(request, *args, **kwargs)



import abc
import hashlib
from collections import OrderedDict

import sentry_sdk
from django.conf import settings
from django.core.cache import cache
from django.core.paginator import InvalidPage, Page, Paginator
from django.db.models.query import QuerySet
from django.utils.functional import cached_property
from rest_framework.exceptions import NotFound
from rest_framework.pagination import PageNumberPagination, _positive_int
from rest_framework.response import Response


class BaseFastPaginator(metaclass=abc.ABCMeta):
    TIMEOUT = getattr(settings, "FAST_PAGINATION_TIMEOUT", 360)
    PREFIX = getattr(settings, "FAST_PAGINATION_PREFIX", "fastpagination")

    @abc.abstractmethod
    def count(self):
        raise NotImplementedError

    @abc.abstractmethod
    def page(self, number):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_page(self, *args, **kwargs):
        raise NotImplementedError


class FastQuerysetPaginator(Paginator, BaseFastPaginator):
    def __init__(self, object_list, per_page, orphans=0, allow_empty_first_page=True):
        super().__init__(object_list, per_page, orphans, allow_empty_first_page)

        try:
            encoded_query = str(object_list.query).encode("utf-8")
            raw_query_key = hashlib.md5(encoded_query, usedforsecurity=False).hexdigest()
            self.cache_pks_key = f"{self.PREFIX}:pks:{raw_query_key}"
            self.cache_count_key = f"{self.PREFIX}:count:{raw_query_key}"
        except Exception:
            encoded_query = None

    @property
    def count(self):
        try:
            result = cache.get(self.cache_count_key)
        except Exception:
            return 0
        if result is None:
            result = self.object_list.count()

            cache.set(self.cache_count_key, result, timeout=self.TIMEOUT)

        return result

    @property
    def pks(self):
        result = cache.get(self.cache_pks_key)
        if result is None:
            result = self.object_list.values_list("pk", flat=True)
            cache.set(self.cache_pks_key, result, timeout=self.TIMEOUT)
        return result

    def page(self, number):
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        if top + self.orphans >= self.count:
            top = self.count
        pks = self.pks[bottom:top]
        object_list = self.object_list.filter(pk__in=pks)
        return self._get_page(object_list, number, self)

    def _get_page(self, *args, **kwargs):
        return FastQuerysetPage(*args, **kwargs)


class FastObjectPaginator(Paginator, BaseFastPaginator):
    def __init__(self, object_list, per_page, orphans=0, allow_empty_first_page=True, cache_key=None):
        if cache_key is None:
            raise ValueError("You should provide cache_key" + "for your results")
        super().__init__(object_list, per_page, orphans, allow_empty_first_page)
        self.cache_count_key = f"{self.PREFIX}:count:{cache_key}"

    def page(self, number):
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        if top + self.orphans >= self.count:
            top = self.count
        object_list = self.object_list[bottom:top]
        return self._get_page(object_list, number, self)

    @property
    def count(self):
        result = cache.get(self.cache_count_key)
        if result is None:
            result = len(self.object_list)
            cache.set(self.cache_count_key, result, timeout=self.TIMEOUT)
        return result

    def _get_page(self, *args, **kwargs):
        return FastObjectPage(*args, **kwargs)


class FastPaginator:
    def __new__(cls, *args, **kwargs):
        object_list = args[0]
        if isinstance(object_list, QuerySet):
            return FastQuerysetPaginator(*args, **kwargs)
        return FastObjectPaginator(*args, **kwargs)


class FastQuerysetPage(Page):
    def __len__(self):
        return len(self.paginator.ids)


class FastObjectPage(Page):
    def __len__(self):
        return len(self.paginator.object_list)


class RecklessPaginator(Paginator):
    @cached_property
    def count(self):
        """Return the total number of objects, across all pages."""
        return 999999999


class StandardResultsSetPagination(PageNumberPagination):
    page_size = 10
    page_size_query_param = "page_size"
    max_page_size = 1000

    def get_page_size(self, request):
        if self.page_size_query_param:
            try:
                return _positive_int(
                    request.query_params[self.page_size_query_param], strict=True, cutoff=self.max_page_size
                )
            except (KeyError, ValueError):
                return self.max_page_size  # return max page size if we receive a negative number

        return self.page_size


class FasterPageNumberPagination(StandardResultsSetPagination):
    django_paginator_class = FastPaginator

    def paginate_queryset(self, queryset, request, view=None):
        page_size = self.get_page_size(request)
        if not page_size:
            return 0
        self._page_size = page_size

        paginator = self.django_paginator_class(queryset, page_size)
        page_number = self.get_page_number(request, paginator)
        self.page_number = page_number

        try:
            self.page = paginator.page(page_number)
        except InvalidPage as exc:
            msg = self.invalid_page_message.format(page_number=page_number, message=str(exc))
            sentry_sdk.capture_exception(Exception(msg))
            raise NotFound("Page not found")

        if paginator.num_pages > 1 and self.template is not None:
            # The browsable API should display pagination controls.
            self.display_page_controls = True

        self.request = request
        return list(self.page)

    def get_next_num(self):
        if not self.page.has_next():
            return None
        return self.page.next_page_number()

    def get_previous_num(self):
        if not self.page.has_previous():
            return None
        return self.page.previous_page_number()

    def get_current_num(self):
        try:
            n = int(self.page_number)
        except Exception:
            n = None
        return n

    def get_paginated_response(self, data):
        return Response(
            OrderedDict(
                [
                    ("count", self.page.paginator.count),
                    ("current", self.get_current_num()),
                    ("next", self.get_next_link()),
                    ("next_num", self.get_next_num()),
                    ("page_size", self._page_size),
                    ("previous", self.get_previous_link()),
                    ("previous_num", self.get_previous_num()),
                    ("results", data),
                ]
            )
        )



"""This module contains utility functions for interacting with celery.

You should avoid importing application specific modules or classes here, this
should only contain generic libraries that can be used for any app.

"""

from functools import partial
from typing import Callable

from django.db import transaction


def queue_task(task: Callable, on_commit: bool = True, countdown: int = 0, *args, **kwargs) -> None | str:
    """Wrapper for calling celery tasks.  By default, tasks will be executed on commit to ensure that data
    has been persisted to the database.

    Examples:
        The most basic call to celery would look like::

           queue_task(your_task, arg1, arg2, kwarg1=val1, kwarg2=val2)

        Additionally, you can provide some arguments that are typically provided to apply_async, such
        as `countdown` and `queue`.

            queue_task(your_task, arg1, arg2, kwarg1=val1, kwarg2=val2, queue="low_priority_queue", countdown=5)

    Args:
        task (Callable): The function to be called as a celery task
        on_commit (bool): Whether to run this task on commit.  Defaults to `True`, and this is the recommended value.
        countdown (int): How long to wait before running this task (matches celery's `countdown` arg). If no `countdown`
            is provided, the task will be immediately queued.
        *args (tuple): Positional arguments to pass to the celery task.  This replaces passing `args=(arg1, arg2, ...)` to
           `apply_async`
        **kwargs (dict): Keyword arguments to pass to the celery task.  This replaces passing `kwargs={"arg1": val1}`

    Keyword Args:
        queue (str): The name of the queue to send the task to.  If no `queue` is provided, the default task
            queue is used.
    """
    task_kwargs = {}
    if queue := kwargs.pop("queue", None):
        task_kwargs["queue"] = queue

    if on_commit:
        transaction.on_commit(partial(task.apply_async, countdown=countdown, args=args, kwargs=kwargs, **task_kwargs))
        return

    return task.apply_async(countdown=countdown, args=args, kwargs=kwargs, **task_kwargs)



import semantic_version
from django.core.serializers.json import DjangoJSONEncoder


class VersionCompatibleEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, semantic_version.Version):
            return str(obj)
        return super().default(obj)



import math
from functools import wraps
from random import randint


def round_up(n, decimals=2):
    multiplier = 10**decimals
    return math.floor(float(n) * multiplier + 0.5) / multiplier


def randN(N):
    min = pow(10, N - 1)
    max = pow(10, N) - 1
    return randint(min, max)


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR", None)
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


def cached_attribute(func):
    cache_name = f"_{func.__name__}"

    @wraps(func)
    def inner(self, *args, **kwargs):
        if hasattr(self, cache_name):
            return getattr(self, cache_name)
        val = func(self, *args, **kwargs)
        setattr(self, cache_name, val)
        return val

    return inner
